#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import warnings
import io
import sys

from pgmpy.readwrite import PomdpXReader, PomdpXWriter

try:
    from lxml import etree
except ImportError:
    try:
        # Python 2.5
        import xml.etree.cElementTree as etree

        print("running with cElementTree on Python 2.5+")
    except ImportError:
        try:
            import xml.etree.ElementTree as etree

            print("running with ElementTree on Python 2.5+")
        except ImportError:
            warnings.warn("Failed to import ElementTree from any known place")


class TestPomdpXReaderString(unittest.TestCase):
    def setUp(self):
        string = """<pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
         <Variable>
              <StateVar vnamePrev="rover_0" vnameCurr="rover_1"
                fullyObs="true">
                  <NumValues>3</NumValues>
              </StateVar>
              <StateVar vnamePrev="rock_0" vnameCurr="rock_1">
                  <ValueEnum>good bad</ValueEnum>
              </StateVar>
              <ObsVar vname="obs_sensor">
                  <ValueEnum>ogood obad</ValueEnum>
              </ObsVar>
              <ActionVar vname="action_rover">
                  <ValueEnum>amw ame ac as</ValueEnum>
              </ActionVar>
              <RewardVar vname="reward_rover" />
         </Variable>
         <InitialStateBelief>
              <CondProb>
                  <Var>rover_0</Var>
                  <Parent>null</Parent>
                  <Parameter type="TBL">
                        <Entry>
                            <Instance> - </Instance>
                            <ProbTable>0.0 1.0 0.0</ProbTable>
                        </Entry>
              </Parameter>
         </CondProb>
         <CondProb>
              <Var>rock_0</Var>
              <Parent>null</Parent>
              <Parameter type="TBL">
                  <Entry>
                      <Instance>-</Instance>
                      <ProbTable>uniform</ProbTable>
                  </Entry>
              </Parameter>
         </CondProb>
      </InitialStateBelief>
      <StateTransitionFunction>
          <CondProb>
              <Var>rover_1</Var>
              <Parent>action_rover rover_0</Parent>
              <Parameter type="TBL">
                  <Entry>
                      <Instance>amw s0 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>amw s1 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ame s0 s1</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ame s1 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ac s0 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>ac s1 s1</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>as s0 s0</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>as s1 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
                  <Entry>
                      <Instance>* s2 s2</Instance>
                      <ProbTable>1.0</ProbTable>
                  </Entry>
           </Parameter>
       </CondProb>
       <CondProb>
           <Var>rock_1</Var>
           <Parent>action_rover rover_0 rock_0</Parent>
           <Parameter>
               <Entry>
                   <Instance>amw * - - </Instance>
                   <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ame * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as * - - </Instance>
                   <ProbTable>identity</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as s0 * - </Instance>
                   <ProbTable>0.0 1.0</ProbTable>
               </Entry>
           </Parameter>
       </CondProb>
   </StateTransitionFunction>
   <ObsFunction>
       <CondProb>
           <Var>obs_sensor</Var>
           <Parent>action_rover rover_1 rock_1</Parent>
           <Parameter type="TBL">
               <Entry>
                   <Instance>amw * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ame * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>as * * - </Instance>
                   <ProbTable>1.0 0.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac s0 - - </Instance>
                   <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
               </Entry>
               <Entry>
                   <Instance>ac s1 - - </Instance>
                   <ProbTable>0.8 0.2 0.2 0.8</ProbTable>
               </Entry>
                <Entry>
                     <Instance>ac s2 * - </Instance>
                     <ProbTable>1.0 0.0</ProbTable>
                </Entry>
            </Parameter>
        </CondProb>
    </ObsFunction>
    <RewardFunction>
        <Func>
            <Var>reward_rover</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="TBL">
                <Entry>
                     <Instance>ame s1 *</Instance>
                     <ValueTable>10</ValueTable>
                </Entry>
                <Entry>
                     <Instance>amw s0 *</Instance>
                     <ValueTable>-100</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s1 *</Instance>
                     <ValueTable>-100</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s0 good</Instance>
                     <ValueTable>10</ValueTable>
                </Entry>
                <Entry>
                     <Instance>as s0 bad</Instance>
                     <ValueTable>-10</ValueTable>
                </Entry>
            </Parameter>
        </Func>
    </RewardFunction>
 </pomdpx>
 """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))

    def test_get_variables(self):
        var_expected = {
            "StateVar": [
                {
                    "vnamePrev": "rover_0",
                    "vnameCurr": "rover_1",
                    "ValueEnum": ["s0", "s1", "s2"],
                    "fullyObs": True,
                },
                {
                    "vnamePrev": "rock_0",
                    "vnameCurr": "rock_1",
                    "fullyObs": False,
                    "ValueEnum": ["good", "bad"],
                },
            ],
            "ObsVar": [{"vname": "obs_sensor", "ValueEnum": ["ogood", "obad"]}],
            "RewardVar": [{"vname": "reward_rover"}],
            "ActionVar": [
                {"vname": "action_rover", "ValueEnum": ["amw", "ame", "ac", "as"]}
            ],
        }
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_variables(), var_expected)
        self.assertEqual(self.reader_file.get_variables(), var_expected)

    def test_get_initial_belief_system(self):
        belief_expected = [
            {
                "Var": "rover_0",
                "Parent": ["null"],
                "Type": "TBL",
                "Parameter": [{"Instance": ["-"], "ProbTable": ["0.0", "1.0", "0.0"]}],
            },
            {
                "Var": "rock_0",
                "Parent": ["null"],
                "Type": "TBL",
                "Parameter": [{"Instance": ["-"], "ProbTable": ["uniform"]}],
            },
        ]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_initial_beliefs(), belief_expected)
        self.assertEqual(self.reader_file.get_initial_beliefs(), belief_expected)

    def test_get_state_transition_function(self):
        state_transition_function_expected = [
            {
                "Var": "rover_1",
                "Parent": ["action_rover", "rover_0"],
                "Type": "TBL",
                "Parameter": [
                    {"Instance": ["amw", "s0", "s2"], "ProbTable": ["1.0"]},
                    {"Instance": ["amw", "s1", "s0"], "ProbTable": ["1.0"]},
                    {"Instance": ["ame", "s0", "s1"], "ProbTable": ["1.0"]},
                    {"Instance": ["ame", "s1", "s2"], "ProbTable": ["1.0"]},
                    {"Instance": ["ac", "s0", "s0"], "ProbTable": ["1.0"]},
                    {"Instance": ["ac", "s1", "s1"], "ProbTable": ["1.0"]},
                    {"Instance": ["as", "s0", "s0"], "ProbTable": ["1.0"]},
                    {"Instance": ["as", "s1", "s2"], "ProbTable": ["1.0"]},
                    {"Instance": ["*", "s2", "s2"], "ProbTable": ["1.0"]},
                ],
            },
            {
                "Var": "rock_1",
                "Parent": ["action_rover", "rover_0", "rock_0"],
                "Type": "TBL",
                "Parameter": [
                    {
                        "Instance": ["amw", "*", "-", "-"],
                        "ProbTable": ["1.0", "0.0", "0.0", "1.0"],
                    },
                    {"Instance": ["ame", "*", "-", "-"], "ProbTable": ["identity"]},
                    {"Instance": ["ac", "*", "-", "-"], "ProbTable": ["identity"]},
                    {"Instance": ["as", "*", "-", "-"], "ProbTable": ["identity"]},
                    {"Instance": ["as", "s0", "*", "-"], "ProbTable": ["0.0", "1.0"]},
                ],
            },
        ]
        self.maxDiff = None
        self.assertEqual(
            self.reader_string.get_state_transition_function(),
            state_transition_function_expected,
        )
        self.assertEqual(
            self.reader_file.get_state_transition_function(),
            state_transition_function_expected,
        )

    def test_obs_function(self):
        obs_function_expected = [
            {
                "Var": "obs_sensor",
                "Parent": ["action_rover", "rover_1", "rock_1"],
                "Type": "TBL",
                "Parameter": [
                    {"Instance": ["amw", "*", "*", "-"], "ProbTable": ["1.0", "0.0"]},
                    {"Instance": ["ame", "*", "*", "-"], "ProbTable": ["1.0", "0.0"]},
                    {"Instance": ["as", "*", "*", "-"], "ProbTable": ["1.0", "0.0"]},
                    {
                        "Instance": ["ac", "s0", "-", "-"],
                        "ProbTable": ["1.0", "0.0", "0.0", "1.0"],
                    },
                    {
                        "Instance": ["ac", "s1", "-", "-"],
                        "ProbTable": ["0.8", "0.2", "0.2", "0.8"],
                    },
                    {"Instance": ["ac", "s2", "*", "-"], "ProbTable": ["1.0", "0.0"]},
                ],
            }
        ]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_obs_function(), obs_function_expected)
        self.assertEqual(self.reader_file.get_obs_function(), obs_function_expected)

    def test_reward_function(self):
        reward_function_expected = [
            {
                "Var": "reward_rover",
                "Parent": ["action_rover", "rover_0", "rock_0"],
                "Type": "TBL",
                "Parameter": [
                    {"Instance": ["ame", "s1", "*"], "ValueTable": ["10"]},
                    {"Instance": ["amw", "s0", "*"], "ValueTable": ["-100"]},
                    {"Instance": ["as", "s1", "*"], "ValueTable": ["-100"]},
                    {"Instance": ["as", "s0", "good"], "ValueTable": ["10"]},
                    {"Instance": ["as", "s0", "bad"], "ValueTable": ["-10"]},
                ],
            }
        ]
        self.maxDiff = None
        self.assertEqual(
            self.reader_string.get_reward_function(), reward_function_expected
        )
        self.assertEqual(
            self.reader_file.get_reward_function(), reward_function_expected
        )

    def test_get_parameter_dd(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <InitialStateBelief>
        <CondProb>
         <Var>rover_0</Var>
  <Parent>null</Parent>
   <Parameter type = "DD">
      <DAG>
         <Node var = "rover_0">
             <Edge val="s0"><Terminal>0.0</Terminal></Edge>
              <Edge val="s1">
                 <Node var = "rock_0">
                     <Edge val = "good">
                       <Terminal>0.5</Terminal>
                  </Edge>
                   <Edge val = "bad">
                      <Terminal>0.5</Terminal>
                   </Edge>
                </Node>
           </Edge>
           <Edge val="s2"><Terminal>0.0</Terminal></Edge>
         </Node>
      </DAG>
  </Parameter>
  </CondProb>
  </InitialStateBelief>
  </pomdpx>
 """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))
        expected_dd_parameter = [
            {
                "Var": "rover_0",
                "Parent": ["null"],
                "Type": "DD",
                "Parameter": {
                    "rover_0": {
                        "s0": "0.0",
                        "s1": {"rock_0": {"good": "0.5", "bad": "0.5"}},
                        "s2": "0.0",
                    }
                },
            }
        ]
        self.maxDiff = None
        self.assertEqual(
            expected_dd_parameter, self.reader_string.get_initial_beliefs()
        )
        self.assertEqual(expected_dd_parameter, self.reader_file.get_initial_beliefs())

    def test_initial_belief_dd(self):
        string = """
    <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <InitialStateBelief>
        <CondProb>
            <Var>rover_0</Var>
            <Parent>null</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="rover_0">
                        <Edge val="s0">
                            <Terminal>0.0</Terminal>
                        </Edge>
                        <Edge val="s1">
                            <SubDAG type="uniform" var="rock_0"/>
                        </Edge>
                        <Edge val="s2">
                            <Terminal>0.0</Terminal>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
    </InitialStateBelief>
    </pomdpx>
    """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))
        expected_belief_dd = [
            {
                "Var": "rover_0",
                "Parent": ["null"],
                "Type": "DD",
                "Parameter": {
                    "rover_0": {
                        "s0": "0.0",
                        "s1": {"type": "uniform", "var": "rock_0"},
                        "s2": "0.0",
                    }
                },
            }
        ]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_initial_beliefs(), expected_belief_dd)
        self.assertEqual(self.reader_file.get_initial_beliefs(), expected_belief_dd)

    def test_reward_function(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <RewardFunction>
        <Func>
            <Var>reward_rover</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Terminal>-100.0</Terminal>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ame">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>10.0</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>0.0</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ac">
                            <Terminal>0.0</Terminal>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <Node var="rock_0">
                                        <Edge val="good">
                                            <Terminal>10</Terminal>
                                        </Edge>
                                        <Edge val="bad">
                                            <Terminal>-10</Terminal>
                                        </Edge>
                                    </Node>
                                </Edge>
                                <Edge val="s1">
                                    <Terminal>-100</Terminal>
                                </Edge>
                                <Edge val="s2">
                                    <Terminal>-100</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </Func>
    </RewardFunction>
    </pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))
        expected_reward_function_dd = [
            {
                "Var": "reward_rover",
                "Parent": ["action_rover", "rover_0", "rock_0"],
                "Type": "DD",
                "Parameter": {
                    "action_rover": {
                        "amw": {"rover_0": {"s0": "-100.0", "s1": "0.0", "s2": "0.0"}},
                        "ame": {"rover_0": {"s0": "0.0", "s1": "10.0", "s2": "0.0"}},
                        "ac": "0.0",
                        "as": {
                            "rover_0": {
                                "s0": {"rock_0": {"good": "10", "bad": "-10"}},
                                "s1": "-100",
                                "s2": "-100",
                            }
                        },
                    }
                },
            }
        ]
        self.maxDiff = None
        self.assertEqual(
            self.reader_string.get_reward_function(), expected_reward_function_dd
        )
        self.assertEqual(
            self.reader_file.get_reward_function(), expected_reward_function_dd
        )

    def test_state_transition_function(self):
        string = """
         <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <StateTransitionFunction>
        <CondProb>
            <Var>rover_1</Var>
            <Parent>action_rover rover_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s0"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ame">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s1"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="ac">
                            <SubDAG type="persistent" var="rover_1"/>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rover_1" val="s0"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="deterministic" var="rover_1" val="s2"/>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
        <CondProb>
            <Var>rock_1</Var>
            <Parent>action_rover rover_0 rock_0</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="ame">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="ac">
                            <SubDAG type="persistent" var="rock_1"/>
                        </Edge>
                        <Edge val="as">
                            <Node var="rover_0">
                                <Edge val="s0">
                                    <SubDAG type="deterministic" var="rock_1" val="bad"/>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="persistent" var="rock_1"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="persistent" var="rock_1"/>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </DAG>
            </Parameter>
        </CondProb>
    </StateTransitionFunction>
</pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))
        expected_state_transition_function = [
            {
                "Var": "rover_1",
                "Parent": ["action_rover", "rover_0"],
                "Type": "DD",
                "Parameter": {
                    "action_rover": {
                        "amw": {
                            "rover_0": {
                                "s0": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                                "s1": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s0",
                                },
                                "s2": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                            }
                        },
                        "ame": {
                            "rover_0": {
                                "s0": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s1",
                                },
                                "s1": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                                "s2": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                            }
                        },
                        "ac": {"type": "persistent", "var": "rover_1"},
                        "as": {
                            "rover_0": {
                                "s0": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s0",
                                },
                                "s1": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                                "s2": {
                                    "type": "deterministic",
                                    "var": "rover_1",
                                    "val": "s2",
                                },
                            }
                        },
                    }
                },
            },
            {
                "Var": "rock_1",
                "Parent": ["action_rover", "rover_0", "rock_0"],
                "Type": "DD",
                "Parameter": {
                    "action_rover": {
                        "amw": {"type": "persistent", "var": "rock_1"},
                        "ame": {"type": "persistent", "var": "rock_1"},
                        "ac": {"type": "persistent", "var": "rock_1"},
                        "as": {
                            "rover_0": {
                                "s0": {
                                    "type": "deterministic",
                                    "var": "rock_1",
                                    "val": "bad",
                                },
                                "s1": {"type": "persistent", "var": "rock_1"},
                                "s2": {"type": "persistent", "var": "rock_1"},
                            }
                        },
                    }
                },
            },
        ]
        self.maxDiff = None
        self.assertEqual(
            self.reader_string.get_state_transition_function(),
            expected_state_transition_function,
        )
        self.assertEqual(
            self.reader_file.get_state_transition_function(),
            expected_state_transition_function,
        )

    def test_obs_function_dd(self):
        string = """
        <pomdpx version="1.0" id="rockSample"
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
      xsi:noNamespaceSchemaLocation="pomdpx.xsd">
         <Description>RockSample problem for map size 1 x 3.
           Rock is at 0, Rover’s initial position is at 1.
           Exit is at 2.
         </Description>
         <Discount>0.95</Discount>
        <ObsFunction>
        <CondProb>
            <Var>obs_sensor</Var>
            <Parent>action_rover rover_1 rock_1</Parent>
            <Parameter type="DD">
                <DAG>
                    <Node var="action_rover">
                        <Edge val="amw">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                        <Edge val="ame">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                        <Edge val="ac">
                            <Node var="rover_1">
                                <Edge val="s0">
                                    <Node var="rock_1">
                                        <Edge val="good">
                                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                                        </Edge>
                                        <Edge val="bad">
                                            <SubDAG type="deterministic" var="obs_sensor" val="obad"/>
                                        </Edge>
                                    </Node>
                                </Edge>
                                <Edge val="s1">
                                    <SubDAG type="template" idref="obs_rock"/>
                                </Edge>
                                <Edge val="s2">
                                    <SubDAG type="template" idref="obs_rock"/>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="as">
                            <SubDAG type="deterministic" var="obs_sensor" val="ogood"/>
                        </Edge>
                    </Node>
                </DAG>
                <SubDAGTemplate id="obs_rock">
                    <Node var="rock_1">
                        <Edge val="good">
                            <Node var="obs_sensor">
                                <Edge val="ogood">
                                    <Terminal>0.8</Terminal>
                                </Edge>
                                <Edge val="obad">
                                    <Terminal>0.2</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                        <Edge val="bad">
                            <Node var="obs_sensor">
                                <Edge val="ogood">
                                    <Terminal>0.2</Terminal>
                                </Edge>
                                <Edge val="obad">
                                    <Terminal>0.8</Terminal>
                                </Edge>
                            </Node>
                        </Edge>
                    </Node>
                </SubDAGTemplate>
            </Parameter>
        </CondProb>
    </ObsFunction>
    </pomdpx>
        """
        self.reader_string = PomdpXReader(string=string)
        self.reader_file = PomdpXReader(path=io.StringIO(string))
        expected_obs_function = [
            {
                "Var": "obs_sensor",
                "Parent": ["action_rover", "rover_1", "rock_1"],
                "Type": "DD",
                "Parameter": {
                    "action_rover": {
                        "amw": {
                            "type": "deterministic",
                            "var": "obs_sensor",
                            "val": "ogood",
                        },
                        "ame": {
                            "type": "deterministic",
                            "var": "obs_sensor",
                            "val": "ogood",
                        },
                        "ac": {
                            "rover_1": {
                                "s0": {
                                    "rock_1": {
                                        "good": {
                                            "type": "deterministic",
                                            "var": "obs_sensor",
                                            "val": "ogood",
                                        },
                                        "bad": {
                                            "type": "deterministic",
                                            "var": "obs_sensor",
                                            "val": "obad",
                                        },
                                    }
                                },
                                "s1": {"type": "template", "idref": "obs_rock"},
                                "s2": {"type": "template", "idref": "obs_rock"},
                            }
                        },
                        "as": {
                            "type": "deterministic",
                            "var": "obs_sensor",
                            "val": "ogood",
                        },
                    },
                    "SubDAGTemplate": {
                        "rock_1": {
                            "good": {"obs_sensor": {"ogood": "0.8", "obad": "0.2"}},
                            "bad": {"obs_sensor": {"ogood": "0.2", "obad": "0.8"}},
                        }
                    },
                    "id": "obs_rock",
                },
            }
        ]
        self.maxDiff = None
        self.assertEqual(self.reader_string.get_obs_function(), expected_obs_function)
        self.assertEqual(self.reader_file.get_obs_function(), expected_obs_function)

    def tearDown(self):
        del self.reader_file
        del self.reader_string


class TestPomdpXWriter(unittest.TestCase):
    def setUp(self):
        self.model_data = {
            "description": "",
            "discount": "0.95",
            "variables": {
                "StateVar": [
                    {
                        "vnamePrev": "rover_0",
                        "vnameCurr": "rover_1",
                        "ValueEnum": ["s0", "s1", "s2"],
                        "fullyObs": True,
                    },
                    {
                        "vnamePrev": "rock_0",
                        "vnameCurr": "rock_1",
                        "fullyObs": False,
                        "ValueEnum": ["good", "bad"],
                    },
                ],
                "ObsVar": [{"vname": "obs_sensor", "ValueEnum": ["ogood", "obad"]}],
                "RewardVar": [{"vname": "reward_rover"}],
                "ActionVar": [
                    {"vname": "action_rover", "ValueEnum": ["amw", "ame", "ac", "as"]}
                ],
            },
            "initial_state_belief": [
                {
                    "Var": "rover_0",
                    "Parent": ["null"],
                    "Type": "TBL",
                    "Parameter": [
                        {"Instance": ["-"], "ProbTable": ["0.0", "1.0", "0.0"]}
                    ],
                },
                {
                    "Var": "rock_0",
                    "Parent": ["null"],
                    "Type": "TBL",
                    "Parameter": [{"Instance": ["-"], "ProbTable": ["uniform"]}],
                },
            ],
            "state_transition_function": [
                {
                    "Var": "rover_1",
                    "Parent": ["action_rover", "rover_0"],
                    "Type": "TBL",
                    "Parameter": [
                        {"Instance": ["amw", "s0", "s2"], "ProbTable": ["1.0"]},
                        {"Instance": ["amw", "s1", "s0"], "ProbTable": ["1.0"]},
                        {"Instance": ["ame", "s0", "s1"], "ProbTable": ["1.0"]},
                        {"Instance": ["ame", "s1", "s2"], "ProbTable": ["1.0"]},
                        {"Instance": ["ac", "s0", "s0"], "ProbTable": ["1.0"]},
                        {"Instance": ["ac", "s1", "s1"], "ProbTable": ["1.0"]},
                        {"Instance": ["as", "s0", "s0"], "ProbTable": ["1.0"]},
                        {"Instance": ["as", "s1", "s2"], "ProbTable": ["1.0"]},
                        {"Instance": ["*", "s2", "s2"], "ProbTable": ["1.0"]},
                    ],
                },
                {
                    "Var": "rock_1",
                    "Parent": ["action_rover", "rover_0", "rock_0"],
                    "Type": "TBL",
                    "Parameter": [
                        {
                            "Instance": ["amw", "*", "-", "-"],
                            "ProbTable": ["1.0", "0.0", "0.0", "1.0"],
                        },
                        {"Instance": ["ame", "*", "-", "-"], "ProbTable": ["identity"]},
                        {"Instance": ["ac", "*", "-", "-"], "ProbTable": ["identity"]},
                        {"Instance": ["as", "*", "-", "-"], "ProbTable": ["identity"]},
                        {
                            "Instance": ["as", "s0", "*", "-"],
                            "ProbTable": ["0.0", "1.0"],
                        },
                    ],
                },
            ],
            "obs_function": [
                {
                    "Var": "obs_sensor",
                    "Parent": ["action_rover", "rover_1", "rock_1"],
                    "Type": "TBL",
                    "Parameter": [
                        {
                            "Instance": ["amw", "*", "*", "-"],
                            "ProbTable": ["1.0", "0.0"],
                        },
                        {
                            "Instance": ["ame", "*", "*", "-"],
                            "ProbTable": ["1.0", "0.0"],
                        },
                        {
                            "Instance": ["as", "*", "*", "-"],
                            "ProbTable": ["1.0", "0.0"],
                        },
                        {
                            "Instance": ["ac", "s0", "-", "-"],
                            "ProbTable": ["1.0", "0.0", "0.0", "1.0"],
                        },
                        {
                            "Instance": ["ac", "s1", "-", "-"],
                            "ProbTable": ["0.8", "0.2", "0.2", "0.8"],
                        },
                        {
                            "Instance": ["ac", "s2", "*", "-"],
                            "ProbTable": ["1.0", "0.0"],
                        },
                    ],
                }
            ],
            "reward_function": [
                {
                    "Var": "reward_rover",
                    "Parent": ["action_rover", "rover_0", "rock_0"],
                    "Type": "TBL",
                    "Parameter": [
                        {"Instance": ["ame", "s1", "*"], "ValueTable": ["10"]},
                        {"Instance": ["amw", "s0", "*"], "ValueTable": ["-100"]},
                        {"Instance": ["as", "s1", "*"], "ValueTable": ["-100"]},
                        {"Instance": ["as", "s0", "good"], "ValueTable": ["10"]},
                        {"Instance": ["as", "s0", "bad"], "ValueTable": ["-10"]},
                    ],
                }
            ],
        }

        self.writer = PomdpXWriter(model_data=self.model_data)

    @unittest.skipIf(sys.version_info[1] >= 8, "xml ordering different in python 3.8")
    def test_variables(self):
        expected_variables = etree.XML(
            """
<Variable>
  <StateVar fullyObs="true" vnameCurr="rover_1" vnamePrev="rover_0">
    <NumValues>3</NumValues>
  </StateVar>
  <StateVar fullyObs="false" vnameCurr="rock_1" vnamePrev="rock_0">
    <ValueEnum>good bad</ValueEnum>
  </StateVar>
  <ObsVar vname="obs_sensor">
    <ValueEnum>ogood obad</ValueEnum>
  </ObsVar>
  <ActionVar vname="action_rover">
    <ValueEnum>amw ame ac as</ValueEnum>
  </ActionVar>
  <RewardVar vname="reward_rover" />
</Variable>"""
        )
        self.maxDiff = None
        self.assertEqual(
            self.writer.get_variables(), etree.tostring(expected_variables)
        )

    def test_add_initial_belief(self):
        expected_belief_xml = etree.XML(
            """
<InitialStateBelief>
  <CondProb>
    <Var>rover_0</Var>
    <Parent>null</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance> - </Instance>
        <ProbTable>0.0 1.0 0.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_0</Var>
    <Parent>null</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance> - </Instance>
        <ProbTable>uniform</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</InitialStateBelief>"""
        )
        self.maxDiff = None
        self.assertEqual(
            str(self.writer.add_initial_belief()),
            str(etree.tostring(expected_belief_xml)),
        )

    def test_add_transition_function(self):
        expected_transition_xml = etree.XML(
            """
<StateTransitionFunction>
  <CondProb>
    <Var>rover_1</Var>
    <Parent>action_rover rover_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw s0 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>amw s1 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame s0 s1</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame s1 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s0 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s1 s1</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s0 s0</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s1 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>* s2 s2</Instance>
        <ProbTable>1.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_1</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw * - - </Instance>
        <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>as * - - </Instance>
        <ProbTable>identity</ProbTable>
      </Entry>
      <Entry>
        <Instance>as s0 * - </Instance>
        <ProbTable>0.0 1.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</StateTransitionFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            self.writer.add_state_transition_function(),
            etree.tostring(expected_transition_xml),
        )

    def test_add_obs_function(self):
        expected_obs_xml = etree.XML(
            """
<ObsFunction>
  <CondProb>
    <Var>obs_sensor</Var>
    <Parent>action_rover rover_1 rock_1</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>amw * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ame * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>as * * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s0 - - </Instance>
        <ProbTable>1.0 0.0 0.0 1.0</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s1 - - </Instance>
        <ProbTable>0.8 0.2 0.2 0.8</ProbTable>
      </Entry>
      <Entry>
        <Instance>ac s2 * - </Instance>
        <ProbTable>1.0 0.0</ProbTable>
      </Entry>
    </Parameter>
  </CondProb>
</ObsFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            self.writer.add_obs_function(), etree.tostring(expected_obs_xml)
        )

    def test_add_reward_function(self):
        expected_reward_xml = etree.XML(
            """
<RewardFunction>
  <Func>
    <Var>reward_rover</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="TBL">
      <Entry>
        <Instance>ame s1 * </Instance>
        <ValueTable>10</ValueTable>
      </Entry>
      <Entry>
        <Instance>amw s0 * </Instance>
        <ValueTable>-100</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s1 * </Instance>
        <ValueTable>-100</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s0 good</Instance>
        <ValueTable>10</ValueTable>
      </Entry>
      <Entry>
        <Instance>as s0 bad</Instance>
        <ValueTable>-10</ValueTable>
      </Entry>
    </Parameter>
  </Func>
</RewardFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            self.writer.add_reward_function(), etree.tostring(expected_reward_xml)
        )

    def test_initial_state_belief_dd(self):
        self.model_data = {
            "initial_state_belief": [
                {
                    "Var": "rover_0",
                    "Parent": ["null"],
                    "Type": "DD",
                    "Parameter": {
                        "rover_0": {
                            "s0": "0.0",
                            "s1": {"type": "uniform", "var": "rock_0"},
                            "s2": "0.0",
                        }
                    },
                }
            ]
        }
        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML(
            """
<InitialStateBelief>
  <CondProb>
    <Var>rover_0</Var>
    <Parent>null</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="rover_0">
          <Edge val="s0">
            <Terminal>0.0</Terminal>
          </Edge>
          <Edge val="s1">
            <SubDAG type="uniform" var="rock_0"/>
          </Edge>
          <Edge val="s2">
            <Terminal>0.0</Terminal>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
</InitialStateBelief>"""
        )
        self.maxDiff = None
        self.assertEqual(self.writer.add_initial_belief(), etree.tostring(expected_xml))

    @unittest.skipIf(sys.version_info[1] >= 8, "xml ordering different in python 3.8")
    def test_state_transition_function_dd(self):
        self.model_data = {
            "state_transition_function": [
                {
                    "Var": "rover_1",
                    "Parent": ["action_rover", "rover_0"],
                    "Type": "DD",
                    "Parameter": {
                        "action_rover": {
                            "amw": {
                                "rover_0": {
                                    "s0": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                    "s1": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s0",
                                    },
                                    "s2": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                }
                            },
                            "ame": {
                                "rover_0": {
                                    "s0": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s1",
                                    },
                                    "s1": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                    "s2": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                }
                            },
                            "ac": {"type": "persistent", "var": "rover_1"},
                            "as": {
                                "rover_0": {
                                    "s0": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s0",
                                    },
                                    "s1": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                    "s2": {
                                        "type": "deterministic",
                                        "var": "rover_1",
                                        "val": "s2",
                                    },
                                }
                            },
                        }
                    },
                },
                {
                    "Var": "rock_1",
                    "Parent": ["action_rover", "rover_0", "rock_0"],
                    "Type": "DD",
                    "Parameter": {
                        "action_rover": {
                            "amw": {"type": "persistent", "var": "rock_1"},
                            "ame": {"type": "persistent", "var": "rock_1"},
                            "ac": {"type": "persistent", "var": "rock_1"},
                            "as": {
                                "rover_0": {
                                    "s0": {
                                        "type": "deterministic",
                                        "var": "rock_1",
                                        "val": "bad",
                                    },
                                    "s1": {"type": "persistent", "var": "rock_1"},
                                    "s2": {"type": "persistent", "var": "rock_1"},
                                }
                            },
                        }
                    },
                },
            ]
        }

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML(
            """
<StateTransitionFunction>
  <CondProb>
    <Var>rover_1</Var>
    <Parent>action_rover rover_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <SubDAG type="persistent" var="rover_1"/>
          </Edge>
          <Edge val="ame">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s1" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="amw">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s0" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="s0" var="rover_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="deterministic" val="s2" var="rover_1"/>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
  <CondProb>
    <Var>rock_1</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="ame">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="amw">
            <SubDAG type="persistent" var="rock_1"/>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <SubDAG type="deterministic" val="bad" var="rock_1"/>
              </Edge>
              <Edge val="s1">
                <SubDAG type="persistent" var="rock_1"/>
              </Edge>
              <Edge val="s2">
                <SubDAG type="persistent" var="rock_1"/>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </CondProb>
</StateTransitionFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            str(self.writer.add_state_transition_function()),
            str(etree.tostring(expected_xml)),
        )

    @unittest.skipIf(
        sys.version_info[1] >= 8, "Ordering of the xml different in python 3.8"
    )
    def test_obs_function_dd(self):
        self.model_data = {
            "obs_function": [
                {
                    "Var": "obs_sensor",
                    "Parent": ["action_rover", "rover_1", "rock_1"],
                    "Type": "DD",
                    "Parameter": {
                        "action_rover": {
                            "amw": {
                                "type": "deterministic",
                                "var": "obs_sensor",
                                "val": "ogood",
                            },
                            "ame": {
                                "type": "deterministic",
                                "var": "obs_sensor",
                                "val": "ogood",
                            },
                            "ac": {
                                "rover_1": {
                                    "s0": {
                                        "rock_1": {
                                            "good": {
                                                "type": "deterministic",
                                                "var": "obs_sensor",
                                                "val": "ogood",
                                            },
                                            "bad": {
                                                "type": "deterministic",
                                                "var": "obs_sensor",
                                                "val": "obad",
                                            },
                                        }
                                    },
                                    "s1": {"type": "template", "idref": "obs_rock"},
                                    "s2": {"type": "template", "idref": "obs_rock"},
                                }
                            },
                            "as": {
                                "type": "deterministic",
                                "var": "obs_sensor",
                                "val": "ogood",
                            },
                        },
                        "SubDAGTemplate": {
                            "rock_1": {
                                "good": {"obs_sensor": {"ogood": "0.8", "obad": "0.2"}},
                                "bad": {"obs_sensor": {"ogood": "0.2", "obad": "0.8"}},
                            }
                        },
                        "id": "obs_rock",
                    },
                }
            ]
        }

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML(
            """
<ObsFunction>
  <CondProb>
    <Var>obs_sensor</Var>
    <Parent>action_rover rover_1 rock_1</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <Node var="rover_1">
              <Edge val="s0">
                <Node var="rock_1">
                  <Edge val="bad">
                    <SubDAG type="deterministic" val="obad" var="obs_sensor" />
                  </Edge>
                  <Edge val="good">
                    <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
                  </Edge>
                </Node>
              </Edge>
              <Edge val="s1">
                <SubDAG idref="obs_rock" type="template"/>
              </Edge>
              <Edge val="s2">
                <SubDAG idref="obs_rock" type="template"/>
              </Edge>
            </Node>
          </Edge>
          <Edge val="ame">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
          </Edge>
          <Edge val="amw">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor" />
          </Edge>
          <Edge val="as">
            <SubDAG type="deterministic" val="ogood" var="obs_sensor"/>
          </Edge>
        </Node>
      </DAG>
      <SubDAGTemplate id="obs_rock">
        <Node var="rock_1">
          <Edge val="bad">
            <Node var="obs_sensor">
              <Edge val="obad">
                <Terminal>0.8</Terminal>
              </Edge>
              <Edge val="ogood">
                <Terminal>0.2</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="good">
            <Node var="obs_sensor">
              <Edge val="obad">
                <Terminal>0.2</Terminal>
              </Edge>
              <Edge val="ogood">
                <Terminal>0.8</Terminal>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </SubDAGTemplate>
    </Parameter>
  </CondProb>
</ObsFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            str(self.writer.add_obs_function()), str(etree.tostring(expected_xml))
        )

    def test_reward_function_dd(self):
        self.model_data = {
            "reward_function": [
                {
                    "Var": "reward_rover",
                    "Parent": ["action_rover", "rover_0", "rock_0"],
                    "Type": "DD",
                    "Parameter": {
                        "action_rover": {
                            "amw": {
                                "rover_0": {"s0": "-100.0", "s1": "0.0", "s2": "0.0"}
                            },
                            "ame": {
                                "rover_0": {"s0": "0.0", "s1": "10.0", "s2": "0.0"}
                            },
                            "ac": "0.0",
                            "as": {
                                "rover_0": {
                                    "s0": {"rock_0": {"good": "10", "bad": "-10"}},
                                    "s1": "-100",
                                    "s2": "-100",
                                }
                            },
                        }
                    },
                }
            ]
        }

        self.writer = PomdpXWriter(model_data=self.model_data)
        expected_xml = etree.XML(
            """
<RewardFunction>
  <Func>
    <Var>reward_rover</Var>
    <Parent>action_rover rover_0 rock_0</Parent>
    <Parameter type="DD">
      <DAG>
        <Node var="action_rover">
          <Edge val="ac">
            <Terminal>0.0</Terminal>
          </Edge>
          <Edge val="ame">
            <Node var="rover_0">
              <Edge val="s0">
                <Terminal>0.0</Terminal>
              </Edge>
              <Edge val="s1">
                <Terminal>10.0</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>0.0</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="amw">
            <Node var="rover_0">
              <Edge val="s0">
                <Terminal>-100.0</Terminal>
              </Edge>
              <Edge val="s1">
                <Terminal>0.0</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>0.0</Terminal>
              </Edge>
            </Node>
          </Edge>
          <Edge val="as">
            <Node var="rover_0">
              <Edge val="s0">
                <Node var="rock_0">
                  <Edge val="bad">
                    <Terminal>-10</Terminal>
                  </Edge>
                  <Edge val="good">
                    <Terminal>10</Terminal>
                  </Edge>
                </Node>
              </Edge>
              <Edge val="s1">
                <Terminal>-100</Terminal>
              </Edge>
              <Edge val="s2">
                <Terminal>-100</Terminal>
              </Edge>
            </Node>
          </Edge>
        </Node>
      </DAG>
    </Parameter>
  </Func>
</RewardFunction>"""
        )
        self.maxDiff = None
        self.assertEqual(
            self.writer.add_reward_function(), etree.tostring(expected_xml)
        )



import unittest

import numpy as np

from pgmpy.factors.distributions import GaussianDistribution as JGD
from pgmpy.sampling import LeapFrog, ModifiedEuler, GradLogPDFGaussian


class TestGradLogPDFGaussian(unittest.TestCase):
    def setUp(self):
        mean = np.array([1, 2, 3, 4])
        covariance = np.array(
            [
                [1, 0.2, 0.4, 0.7],
                [0.2, 2, 0.5, 0.8],
                [0.4, 0.5, 3, 0.6],
                [0.7, 0.8, 0.6, 4],
            ]
        )
        self.test_model = JGD(["x", "y", "z", "t"], mean, covariance)
        self.test_gradient = GradLogPDFGaussian([0, 0, 0, 0], self.test_model)

    def test_error(self):
        with self.assertRaises(TypeError):
            GradLogPDFGaussian(1, self.test_model)
        with self.assertRaises(ValueError):
            GradLogPDFGaussian([1, 1], self.test_model)

    def test_gradient(self):
        grad, log = self.test_gradient.get_gradient_log_pdf()
        np.testing.assert_almost_equal(
            grad, np.array([0.05436475, 0.49454937, 0.75465073, 0.77837868])
        )
        np.testing.assert_almost_equal(log, -3.21046521505)


class TestLeapFrog(unittest.TestCase):
    def setUp(self):
        mean = np.array([-1, 1, -1])
        covariance = np.array([[1, 0.6, 0.5], [0.6, 2, 0.3], [0.5, 0.3, 1]])
        self.test_model = JGD(["x", "y", "z"], mean, covariance)
        position = [0, 0, 0]
        momentum = [-1, -1, -1]
        self.test_with_grad_log = LeapFrog(
            model=self.test_model,
            position=position,
            momentum=momentum,
            stepsize=0.3,
            grad_log_pdf=GradLogPDFGaussian,
            grad_log_position=None,
        )
        grad_log_position, _ = GradLogPDFGaussian(
            position, self.test_model
        ).get_gradient_log_pdf()
        self.test_without_grad_log = LeapFrog(
            model=self.test_model,
            position=position,
            momentum=momentum,
            stepsize=0.4,
            grad_log_pdf=GradLogPDFGaussian,
            grad_log_position=grad_log_position,
        )

    def test_errors(self):
        with self.assertRaises(TypeError):
            LeapFrog(
                model=self.test_model,
                position=1,
                momentum=[1, 1],
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
            )
        with self.assertRaises(TypeError):
            LeapFrog(
                model=self.test_model,
                position=[1, 1],
                momentum=1,
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
            )
        with self.assertRaises(ValueError):
            LeapFrog(
                model=self.test_model,
                position=[1, 1],
                momentum=[1],
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
            )
        with self.assertRaises(TypeError):
            LeapFrog(
                model=self.test_model,
                position=[1],
                momentum=[1],
                stepsize=0.1,
                grad_log_pdf=1,
            )
        with self.assertRaises(ValueError):
            LeapFrog(
                model=self.test_model,
                position=[1, 1],
                momentum=[1, 1],
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
            )
        with self.assertRaises(TypeError):
            LeapFrog(
                model=self.test_model,
                position=[1, 1, 1],
                momentum=[1, 1, 1],
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
                grad_log_position=1,
            )
        with self.assertRaises(ValueError):
            LeapFrog(
                model=self.test_model,
                position=[1, 1, 1],
                momentum=[1, 1, 1],
                stepsize=0.1,
                grad_log_pdf=GradLogPDFGaussian,
                grad_log_position=[1, 1],
            )

    def test_leapfrog_methods(self):
        new_pos, new_momentum, new_grad = self.test_with_grad_log.get_proposed_values()
        np.testing.assert_almost_equal(
            new_pos, np.array([-0.35634146, -0.25609756, -0.33])
        )
        np.testing.assert_almost_equal(
            new_momentum, np.array([-1.3396624, -0.70344884, -1.16963415])
        )
        np.testing.assert_almost_equal(
            new_grad, np.array([-1.0123835, 1.00139798, -0.46422764])
        )
        (
            new_pos,
            new_momentum,
            new_grad,
        ) = self.test_without_grad_log.get_proposed_values()
        np.testing.assert_almost_equal(
            new_pos, np.array([-0.5001626, -0.32195122, -0.45333333])
        )
        np.testing.assert_almost_equal(
            new_momentum, np.array([-1.42947981, -0.60709102, -1.21246612])
        )
        np.testing.assert_almost_equal(
            new_grad, np.array([-0.89536651, 0.98893516, -0.39566396])
        )

    def tearDown(self):
        del self.test_model
        del self.test_with_grad_log
        del self.test_without_grad_log


class TestModifiedEuler(unittest.TestCase):
    def setUp(self):
        mean = np.array([0, 0])
        covariance = np.array([[-1, 0.8], [0.8, 3]])
        self.test_model = JGD(["x", "y"], mean, covariance)
        position = [0, 0]
        momentum = [-2, 1]
        self.test_with_grad_log = ModifiedEuler(
            model=self.test_model,
            position=position,
            momentum=momentum,
            stepsize=0.5,
            grad_log_pdf=GradLogPDFGaussian,
            grad_log_position=None,
        )
        grad_log_position, _ = GradLogPDFGaussian(
            position, self.test_model
        ).get_gradient_log_pdf()
        self.test_without_grad_log = ModifiedEuler(
            model=self.test_model,
            position=position,
            momentum=momentum,
            stepsize=0.3,
            grad_log_pdf=GradLogPDFGaussian,
            grad_log_position=grad_log_position,
        )

    def test_modified_euler_methods(self):
        new_pos, new_momentum, new_grad = self.test_with_grad_log.get_proposed_values()
        np.testing.assert_almost_equal(new_pos, np.array([-1.0, 0.5]))
        np.testing.assert_almost_equal(new_momentum, np.array([-2.0, 1.0]))
        np.testing.assert_almost_equal(new_grad, np.array([-0.93406593, 0.08241758]))
        (
            new_pos,
            new_momentum,
            new_grad,
        ) = self.test_without_grad_log.get_proposed_values()
        np.testing.assert_almost_equal(new_pos, np.array([-0.6, 0.3]))
        np.testing.assert_almost_equal(new_momentum, np.array([-2.0, 1.0]))
        np.testing.assert_almost_equal(new_grad, np.array([-0.56043956, 0.04945055]))

    def tearDown(self):
        del self.test_model
        del self.test_with_grad_log
        del self.test_without_grad_log



import unittest

import networkx as nx
import numpy as np
import pandas as pd

from pgmpy.estimators import PC
from pgmpy.independencies import Independencies
from pgmpy.models import BayesianNetwork
from pgmpy.sampling import BayesianModelSampling
from pgmpy.utils import get_example_model


# This class tests examples from: Le, Thuc, et al. "A fast PC algorithm for
# high dimensional causal discovery with multi-core PCs." IEEE/ACM transactions
# on computational biology and bioinformatics (2016).
class TestPCFakeCITest(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.fake_data = pd.DataFrame(
            np.random.random((1000, 4)), columns=["A", "B", "C", "D"]
        )
        self.estimator = PC(self.fake_data)

    @staticmethod
    def fake_ci_t(X, Y, Z=[], **kwargs):
        """
        A mock CI testing function which gives False for every condition
        except for the following:
            1. B \u27C2 C
            2. B \u27C2 D
            3. C \u27C2 D
            4. A \u27C2 B | C
            5. A \u27C2 C | B
        """
        Z = list(Z)
        if X == "B":
            if Y == "C" or Y == "D":
                return True
            elif Y == "A" and Z == ["C"]:
                return True
        elif X == "C" and Y == "D" and Z == []:
            return True
        elif X == "D" and Y == "C" and Z == []:
            return True
        elif Y == "B":
            if X == "C" or X == "D":
                return True
            elif X == "A" and Z == ["C"]:
                return True
        elif X == "A" and Y == "C" and Z == ["B"]:
            return True
        elif X == "C" and Y == "A" and Z == ["B"]:
            return True
        return False

    def test_build_skeleton_orig(self):
        skel, sep_set = self.estimator.build_skeleton(
            ci_test=TestPCFakeCITest.fake_ci_t, variant="orig"
        )
        expected_edges = {("A", "C"), ("A", "D")}
        for u, v in skel.edges():
            self.assertTrue(((u, v) in expected_edges) or ((v, u) in expected_edges))

        skel, sep_set = self.estimator.build_skeleton(
            ci_test=TestPCFakeCITest.fake_ci_t, max_cond_vars=0, variant="orig"
        )
        expected_edges = {("A", "B"), ("A", "C"), ("A", "D")}
        for u, v in skel.edges():
            self.assertTrue(((u, v) in expected_edges) or ((v, u) in expected_edges))

    def test_build_skeleton_stable(self):
        skel, sep_set = self.estimator.build_skeleton(
            ci_test=TestPCFakeCITest.fake_ci_t, variant="stable"
        )
        expected_edges = {("A", "C"), ("A", "D")}
        for u, v in skel.edges():
            self.assertTrue(((u, v) in expected_edges) or ((v, u) in expected_edges))

        skel, sep_set = self.estimator.build_skeleton(
            ci_test=TestPCFakeCITest.fake_ci_t, max_cond_vars=0, variant="stable"
        )
        expected_edges = {("A", "B"), ("A", "C"), ("A", "D")}
        for u, v in skel.edges():
            self.assertTrue(((u, v) in expected_edges) or ((v, u) in expected_edges))


class TestPCEstimatorFromIndependencies(unittest.TestCase):
    def test_build_skeleton_from_ind(self):
        # Specify a set of independencies
        for variant in ["orig", "stable", "parallel"]:
            ind = Independencies(["B", "C"], ["A", ["B", "C"], "D"])
            ind = ind.closure()
            estimator = PC(independencies=ind)
            skel, sep_sets = estimator.estimate(
                variant=variant,
                ci_test="independence_match",
                return_type="skeleton",
                show_progress=False,
            )

            expected_edges = {("A", "D"), ("B", "D"), ("C", "D")}
            expected_sepsets = {
                frozenset(("A", "C")): tuple(),
                frozenset(("A", "B")): tuple(),
                frozenset(("C", "B")): tuple(),
            }
            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges) or ((v, u) in expected_edges)
                )
            self.assertEqual(sep_sets, expected_sepsets)

            # Generate independencies from a model.
            model = BayesianNetwork([("A", "C"), ("B", "C"), ("B", "D"), ("C", "E")])
            estimator = PC(independencies=model.get_independencies())
            skel, sep_sets = estimator.estimate(
                variant=variant,
                ci_test="independence_match",
                return_type="skeleton",
                show_progress=False,
            )

            expected_edges = model.edges()
            expected_sepsets1 = {
                frozenset(("D", "C")): ("B",),
                frozenset(("E", "B")): ("C",),
                frozenset(("A", "D")): tuple(),
                frozenset(("E", "D")): ("C",),
                frozenset(("E", "A")): ("C",),
                frozenset(("A", "B")): tuple(),
            }
            expected_sepsets2 = {
                frozenset(("D", "C")): ("B",),
                frozenset(("E", "B")): ("C",),
                frozenset(("A", "D")): tuple(),
                frozenset(("E", "D")): ("B",),
                frozenset(("E", "A")): ("C",),
                frozenset(("A", "B")): tuple(),
            }
            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges) or ((v, u) in expected_edges)
                )

            self.assertTrue(
                (sep_sets == expected_sepsets1) or (sep_sets == expected_sepsets2)
            )

    def test_skeleton_to_pdag(self):
        skel = nx.Graph([("A", "D"), ("A", "C"), ("B", "C")])
        sep_sets = {
            frozenset({"D", "C"}): ("A",),
            frozenset({"A", "B"}): tuple(),
            frozenset({"D", "B"}): ("A",),
        }
        pdag = PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets)
        self.assertSetEqual(
            set(pdag.edges()), set([("B", "C"), ("A", "D"), ("A", "C"), ("D", "A")])
        )

        skel = nx.Graph([("A", "B"), ("A", "C")])
        sep_sets = {frozenset({"B", "C"}): ()}
        self.assertSetEqual(
            set(PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets).edges()),
            set([("B", "A"), ("C", "A")]),
        )

        sep_sets = {frozenset({"B", "C"}): ("A",)}
        pdag = PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets)
        self.assertSetEqual(
            set(PC.skeleton_to_pdag(skel, sep_sets).edges()),
            set([("A", "B"), ("B", "A"), ("A", "C"), ("C", "A")]),
        )

        skel = nx.Graph([("A", "C"), ("B", "C"), ("C", "D")])
        sep_sets = {
            frozenset({"A", "B"}): tuple(),
            frozenset({"A", "D"}): ("C",),
            frozenset({"B", "D"}): ("C",),
        }
        pdag = PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets)
        self.assertSetEqual(
            set(pdag.edges()), set([("A", "C"), ("B", "C"), ("C", "D")])
        )

        skel = nx.Graph([("A", "B"), ("A", "C"), ("B", "C"), ("B", "D")])
        sep_sets = {frozenset({"A", "D"}): tuple(), frozenset({"C", "D"}): ("A", "B")}
        pdag = PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets)
        self.assertSetEqual(
            set(pdag.edges()), set([("A", "B"), ("B", "C"), ("A", "C"), ("D", "B")])
        )

        skel = nx.Graph([("A", "B"), ("B", "C"), ("A", "D"), ("B", "D"), ("C", "D")])
        sep_sets = {frozenset({"A", "C"}): ("B",)}
        pdag = PC.skeleton_to_pdag(skeleton=skel, separating_sets=sep_sets)
        self.assertSetEqual(
            set(pdag.edges()),
            set(
                [
                    ("A", "B"),
                    ("B", "A"),
                    ("B", "C"),
                    ("C", "B"),
                    ("A", "D"),
                    ("B", "D"),
                    ("C", "D"),
                ]
            ),
        )

    def test_estimate_dag(self):
        for variant in ["orig", "stable", "parallel"]:
            ind = Independencies(["B", "C"], ["A", ["B", "C"], "D"])
            ind = ind.closure()
            estimator = PC(independencies=ind)
            model = estimator.estimate(
                variant="orig",
                ci_test="independence_match",
                return_type="dag",
                show_progress=False,
            )
            expected_edges = {("B", "D"), ("A", "D"), ("C", "D")}
            self.assertEqual(model.edges(), expected_edges)

            model = BayesianNetwork([("A", "C"), ("B", "C"), ("B", "D"), ("C", "E")])
            estimator = PC(independencies=model.get_independencies())
            estimated_model = estimator.estimate(
                variant="orig",
                ci_test="independence_match",
                return_type="dag",
                show_progress=False,
            )
            expected_edges_1 = set(model.edges())
            expected_edges_2 = {("B", "C"), ("A", "C"), ("C", "E"), ("D", "B")}
            self.assertTrue(
                (set(estimated_model.edges()) == expected_edges_1)
                or (set(estimated_model.edges()) == expected_edges_2)
            )


class TestPCEstimatorFromDiscreteData(unittest.TestCase):
    def test_build_skeleton(self):
        for variant in ["orig", "stable", "parallel"]:
            # Fake dataset no: 1
            np.random.seed(42)
            data = pd.DataFrame(
                np.random.randint(0, 2, size=(10000, 5)), columns=list("ABCDE")
            )
            data["F"] = data["A"] + data["B"] + data["C"]
            est = PC(data=data)
            skel, sep_sets = est.estimate(
                variant=variant,
                ci_test="chi_square",
                return_type="skeleton",
                significance_level=0.005,
                show_progress=False,
            )
            expected_edges = {("A", "F"), ("B", "F"), ("C", "F")}
            expected_sepsets = {
                frozenset(("D", "F")): tuple(),
                frozenset(("D", "B")): tuple(),
                frozenset(("A", "C")): tuple(),
                frozenset(("D", "E")): tuple(),
                frozenset(("E", "F")): tuple(),
                frozenset(("E", "C")): tuple(),
                frozenset(("E", "B")): tuple(),
                frozenset(("D", "C")): tuple(),
                frozenset(("A", "B")): tuple(),
                frozenset(("A", "E")): tuple(),
                frozenset(("B", "C")): tuple(),
                frozenset(("A", "D")): tuple(),
            }
            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges) or ((v, u) in expected_edges)
                )
            self.assertEqual(sep_sets, expected_sepsets)

            # Fake dataset no: 2 Expected structure X <- Z -> Y
            def fake_ci(X, Y, Z=tuple(), **kwargs):
                if X == "X" and Y == "Y" and Z == ("Z",):
                    return True
                elif X == "Y" and Y == "X" and Z == ("Z",):
                    return True
                else:
                    return False

            np.random.seed(42)
            fake_data = pd.DataFrame(
                np.random.randint(low=0, high=2, size=(10000, 3)),
                columns=["X", "Y", "Z"],
            )
            est = PC(data=fake_data)
            skel, sep_sets = est.estimate(
                variant=variant,
                ci_test=fake_ci,
                return_type="skeleton",
                show_progress=False,
            )
            expected_edges = {("X", "Z"), ("Y", "Z")}
            expected_sepsets = {frozenset(("X", "Y")): ("Z",)}
            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges) or ((v, u) in expected_edges)
                )
            self.assertEqual(sep_sets, expected_sepsets)

    def test_build_dag(self):
        for variant in ["orig", "stable", "parallel"]:
            np.random.seed(42)
            data = pd.DataFrame(
                np.random.randint(0, 3, size=(10000, 3)), columns=list("XYZ")
            )
            data["sum"] = data.sum(axis=1)
            est = PC(data=data)
            dag = est.estimate(
                variant=variant,
                ci_test="chi_square",
                return_type="dag",
                significance_level=0.001,
                show_progress=False,
            )
            expected_edges = {("Z", "sum"), ("X", "sum"), ("Y", "sum")}
            self.assertEqual(set(dag.edges()), expected_edges)


class TestPCEstimatorFromContinuousData(unittest.TestCase):
    def test_build_skeleton(self):
        for variant in ["orig", "stable", "parallel"]:
            # Fake dataset no: 1
            np.random.seed(42)
            data = pd.DataFrame(np.random.randn(10000, 5), columns=list("ABCDE"))
            data["F"] = data["A"] + data["B"] + data["C"]
            est = PC(data=data)
            skel, sep_sets = est.estimate(
                variant=variant,
                ci_test="pearsonr",
                return_type="skeleton",
                show_progress=False,
            )
            expected_edges = {("A", "F"), ("B", "F"), ("C", "F")}
            expected_edges_stable = {("A", "F"), ("B", "C"), ("B", "F"), ("C", "F")}
            expected_sepsets = {
                frozenset(("D", "F")): tuple(),
                frozenset(("D", "B")): tuple(),
                frozenset(("A", "C")): tuple(),
                frozenset(("D", "E")): tuple(),
                frozenset(("E", "F")): tuple(),
                frozenset(("E", "C")): tuple(),
                frozenset(("E", "B")): tuple(),
                frozenset(("D", "C")): tuple(),
                frozenset(("A", "B")): tuple(),
                frozenset(("A", "E")): tuple(),
                frozenset(("B", "C")): tuple(),
                frozenset(("A", "D")): tuple(),
                # This one is only for stable version.
                frozenset(("C", "B")): tuple(),
            }
            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges_stable)
                    or ((v, u) in expected_edges_stable)
                )

            for key, value in sep_sets.items():
                self.assertEqual(sep_sets[key], expected_sepsets[key])

            # Fake dataset no: 2. Expected model structure X <- Z -> Y
            def fake_ci(X, Y, Z=tuple(), **kwargs):
                if X == "X" and Y == "Y" and Z == ("Z",):
                    return True
                elif X == "Y" and Y == "X" and Z == ("Z",):
                    return True
                else:
                    return False

            np.random.seed(42)
            data = pd.DataFrame(np.random.randn(10000, 3), columns=list("XYZ"))
            est = PC(data=data)
            skel, sep_sets = est.estimate(
                variant=variant,
                ci_test=fake_ci,
                return_type="skeleton",
                show_progress=False,
            )
            expected_edges = {("X", "Z"), ("Y", "Z")}
            expected_sepsets = {frozenset(("X", "Y")): ("Z",)}

            for u, v in skel.edges():
                self.assertTrue(
                    ((u, v) in expected_edges) or ((v, u) in expected_edges)
                )
            self.assertEqual(sep_sets, expected_sepsets)

    def test_build_dag(self):
        for variant in ["orig", "stable", "parallel"]:
            np.random.seed(42)
            data = pd.DataFrame(np.random.randn(10000, 3), columns=list("XYZ"))
            data["sum"] = data.sum(axis=1)
            est = PC(data=data)
            dag = est.estimate(
                variant=variant,
                ci_test="pearsonr",
                return_type="dag",
                show_progress=False,
            )

            expected_edges = {("Z", "sum"), ("X", "sum"), ("Y", "sum")}
            self.assertEqual(set(dag.edges()), expected_edges)


class TestPCRealModels(unittest.TestCase):
    def test_pc_alarm(self):
        alarm_model = get_example_model("alarm")
        data = BayesianModelSampling(alarm_model).forward_sample(size=int(1e3), seed=42)
        est = PC(data)
        dag = est.estimate(variant="stable", max_cond_vars=1, show_progress=False)

    def test_pc_asia(self):
        asia_model = get_example_model("asia")
        data = BayesianModelSampling(asia_model).forward_sample(size=int(1e3), seed=42)
        est = PC(data)
        dag = est.estimate(variant="stable", max_cond_vars=1, show_progress=False)



"""
Contains simple check functions
"""
import numpy as np


def _check_1d_array_object(parameter, name_param):
    """
    Checks whether given parameter is a 1d array like object, and returns a numpy array object
    """
    if isinstance(parameter, (np.ndarray, list, tuple, np.matrix)):
        parameter = np.array(parameter)
        if parameter.ndim != 1:
            raise TypeError(f"{name_param} should be a 1d array type object")
    else:
        raise TypeError(f"{name_param} should be a 1d array type object")

    return parameter


def _check_length_equal(param_1, param_2, name_param_1, name_param_2):
    """
    Raises an error when the length of given two arguments is not equal
    """
    if len(param_1) != len(param_2):
        raise ValueError(
            f"Length of {name_param_1} must be same as Length of {name_param_2}"
        )



#!/usr/bin/env python

from itertools import combinations

import networkx as nx
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.metrics import (
    adjusted_mutual_info_score,
    mutual_info_score,
    normalized_mutual_info_score,
)
from tqdm.auto import tqdm

from pgmpy.base import DAG
from pgmpy.estimators import StructureEstimator
from pgmpy.global_vars import SHOW_PROGRESS


class TreeSearch(StructureEstimator):
    def __init__(self, data, root_node=None, n_jobs=-1, **kwargs):
        """
        Search class for learning tree related graph structure. The algorithms
        supported are Chow-Liu and Tree-augmented naive bayes (TAN).

        Chow-Liu constructs the maximum-weight spanning tree with mutual information
        score as edge weights.

        TAN is an extension of Naive Bayes classifier to allow a tree structure over
        the independent variables to account for interaction.

        Parameters
        ----------
        data: pandas.DataFrame object
            dataframe object where each column represents one variable.

        root_node: str, int, or any hashable python object, default is None.
            The root node of the tree structure. If None then root node is auto-picked
            as the node with the highest sum of edge weights.

        n_jobs: int (default: -1)
            Number of jobs to run in parallel. `-1` means use all processors.

        References
        ----------
        [1] Chow, C. K.; Liu, C.N. (1968), "Approximating discrete probability
            distributions with dependence trees", IEEE Transactions on Information
            Theory, IT-14 (3): 462–467

        [2] Friedman N, Geiger D and Goldszmidt M (1997). Bayesian network classifiers.
            Machine Learning 29: 131–163
        """
        if root_node is not None and root_node not in data.columns:
            raise ValueError(f"Root node: {root_node} not found in data columns.")

        self.data = data
        self.root_node = root_node
        self.n_jobs = n_jobs

        super(TreeSearch, self).__init__(data, **kwargs)

    def estimate(
        self,
        estimator_type="chow-liu",
        class_node=None,
        edge_weights_fn="mutual_info",
        show_progress=True,
    ):
        """
        Estimate the `DAG` structure that fits best to the given data set without
        parametrization.

        Parameters
        ----------
        estimator_type: str (chow-liu | tan)
            The algorithm to use for estimating the DAG.

        class_node: string, int or any hashable python object. (optional)
            Needed only if estimator_type = 'tan'. In the estimated DAG, there would be
            edges from class_node to each of the feature variables.

        edge_weights_fn: str or function (default: mutual info)
            Method to use for computing edge weights. By default, Mutual Info Score is
            used.

        show_progress: boolean
            If True, shows a progress bar for the running algorithm.

        Returns
        -------
        Estimated Model: pgmpy.base.DAG
            The estimated model structure.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import networkx as nx
        >>> import matplotlib.pyplot as plt
        >>> from pgmpy.estimators import TreeSearch
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 5)),
        ...                       columns=['A', 'B', 'C', 'D', 'E'])
        >>> est = TreeSearch(values, root_node='B')
        >>> model = est.estimate(estimator_type='chow-liu')
        >>> nx.draw_circular(model, with_labels=True, arrowsize=20, arrowstyle='fancy',
        ...                  alpha=0.3)
        >>> plt.show()
        >>> est = TreeSearch(values)
        >>> model = est.estimate(estimator_type='chow-liu')
        >>> nx.draw_circular(model, with_labels=True, arrowsize=20, arrowstyle='fancy',
        ...                  alpha=0.3)
        >>> plt.show()
        >>> est = TreeSearch(values, root_node='B')
        >>> model = est.estimate(estimator_type='tan', class_node='A')
        >>> nx.draw_circular(model, with_labels=True, arrowsize=20, arrowstyle='fancy',
        ...                  alpha=0.3)
        >>> plt.show()
        >>> est = TreeSearch(values)
        >>> model = est.estimate(estimator_type='tan')
        >>> nx.draw_circular(model, with_labels=True, arrowsize=20, arrowstyle='fancy',
        ...                  alpha=0.3)
        >>> plt.show()
        """
        # Step 1. Argument checks
        # Step 1.1: Only chow-liu and tan allowed as estimator type.
        if estimator_type not in {"chow-liu", "tan"}:
            raise ValueError(
                f"Invalid estimator_type. Expected either chow-liu or tan. Got: {estimator_type}"
            )

        # Step 1.2: If estimator_type=tan, class_node must be specified
        if estimator_type == "tan" and class_node is None:
            raise ValueError(
                f"class_node argument must be specified for estimator_type='tan'"
            )
        if estimator_type == "tan" and class_node not in self.data.columns:
            raise ValueError(f"Class node: {class_node} not found in data columns")

        # Step 1.3: If root_node isn't specified, get the node with the highest score.
        weights_computed = False
        if self.root_node is None:
            weights = TreeSearch._get_weights(
                self.data, edge_weights_fn, self.n_jobs, show_progress
            )
            weights_computed = True
            sum_weights = weights.sum(axis=0)
            maxw_idx = np.argsort(sum_weights)[::-1]
            self.root_node = self.data.columns[maxw_idx[0]]

        # Step 2. Compute all edge weights.
        if estimator_type == "chow-liu":
            if not weights_computed:
                weights = TreeSearch._get_weights(
                    self.data, edge_weights_fn, self.n_jobs, show_progress
                )
        else:
            weights = TreeSearch._get_conditional_weights(
                self.data, class_node, edge_weights_fn, self.n_jobs, show_progress
            )

        # Step 3: If estimator_type = "chow-liu", estimate the DAG and return.
        if estimator_type == "chow-liu":
            return TreeSearch._create_tree_and_dag(
                weights, self.data.columns, self.root_node
            )

        # Step 4: If estimator_type = "tan":
        elif estimator_type == "tan":

            # Step 4.1: Checks root_node != class_node
            if self.root_node == class_node:
                raise ValueError(
                    f"Root node: {self.root_node} and class node: {class_node} are identical"
                )

            # Step 4.2: Construct chow-liu DAG on {data.columns - class_node}
            class_node_idx = np.where(self.data.columns == class_node)[0][0]
            weights = np.delete(weights, class_node_idx, axis=0)
            weights = np.delete(weights, class_node_idx, axis=1)
            reduced_columns = np.delete(self.data.columns, class_node_idx)
            D = TreeSearch._create_tree_and_dag(
                weights, reduced_columns, self.root_node
            )

            # Step 4.3: Add edges from class_node to all other nodes.
            D.add_edges_from([(class_node, node) for node in reduced_columns])
            return D

    @staticmethod
    def _get_weights(
        data, edge_weights_fn="mutual_info", n_jobs=-1, show_progress=True
    ):
        """
        Helper function to Chow-Liu algorithm for estimating tree structure from given data. Refer to
        pgmpy.estimators.TreeSearch for more details. This function returns the edge weights matrix.

        Parameters
        ----------
        data: pandas.DataFrame object
            dataframe object where each column represents one variable.

        edge_weights_fn: str or function (default: mutual_info)
            Method to use for computing edge weights. Options are:
                1. 'mutual_info': Mutual Information Score.
                2. 'adjusted_mutual_info': Adjusted Mutual Information Score.
                3. 'normalized_mutual_info': Normalized Mutual Information Score.
                4. function(array[n_samples,], array[n_samples,]): Custom function.

        n_jobs: int (default: -1)
            Number of jobs to run in parallel. `-1` means use all processors.

        show_progress: boolean
            If True, shows a progress bar for the running algorithm.

        Returns
        -------
        weights: numpy 2D array, shape = (n_columns, n_columns)
            symmetric matrix where each element represents an edge weight.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.estimators import TreeSearch
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 5)),
        ...                       columns=['A', 'B', 'C', 'D', 'E'])
        >>> est = TreeSearch(values, root_node='B')
        >>> model = est.estimate(estimator_type='chow-liu')
        """
        # Step 0: Check for edge weight computation method
        if edge_weights_fn == "mutual_info":
            edge_weights_fn = mutual_info_score
        elif edge_weights_fn == "adjusted_mutual_info":
            edge_weights_fn = adjusted_mutual_info_score
        elif edge_weights_fn == "normalized_mutual_info":
            edge_weights_fn = normalized_mutual_info_score
        elif not callable(edge_weights_fn):
            raise ValueError(
                f"edge_weights_fn should either be 'mutual_info', 'adjusted_mutual_info', "
                f"'normalized_mutual_info'or a function of form fun(array, array). Got: f{edge_weights_fn}"
            )

        # Step 1: Compute edge weights for a fully connected graph.
        n_vars = len(data.columns)
        pbar = combinations(data.columns, 2)
        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(pbar, total=(n_vars * (n_vars - 1) / 2), desc="Building tree")

        vals = Parallel(n_jobs=n_jobs, prefer="threads")(
            delayed(edge_weights_fn)(data.loc[:, u], data.loc[:, v]) for u, v in pbar
        )
        weights = np.zeros((n_vars, n_vars))
        indices = np.triu_indices(n_vars, k=1)
        weights[indices] = vals
        weights.T[indices] = vals

        return weights

    @staticmethod
    def _get_conditional_weights(
        data, class_node, edge_weights_fn="mutual_info", n_jobs=-1, show_progress=True
    ):
        """
        Helper function to TAN (Tree Augmented Naive Bayes) algorithm for
        estimating tree structure from given data. Refer to
        pgmpy.estimators.TreeSearch for more details. This function returns the
        edge weights matrix.

        Parameters
        ----------
        data: pandas.DataFrame object
            dataframe object where each column represents one variable.

        class_node: str
            The class node for TAN. The edge weight is computed as I(X, Y | class_node).

        edge_weights_fn: str or function (default: mutual_info)
            Method to use for computing edge weights. Options are:
                1. 'mutual_info': Mutual Information Score.
                2. 'adjusted_mutual_info': Adjusted Mutual Information Score.
                3. 'normalized_mutual_info': Normalized Mutual Information Score.
                4. function(array[n_samples,], array[n_samples,]): Custom function.

        n_jobs: int (default: -1)
            Number of jobs to run in parallel. `-1` means use all processors.

        show_progress: boolean
            If True, shows a progress bar for the running algorithm.

        Returns
        -------
        weights: numpy 2D array, shape = (n_columns, n_columns)
            symmetric matrix where each element represents an edge weight.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.estimators import TreeSearch
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 5)),
        ...                       columns=['A', 'B', 'C', 'D', 'E'])
        >>> est = TreeSearch(values, root_node='B')
        >>> model = est.estimate(estimator_type='tan')
        """
        # Step 0: Check for edge weight computation method
        if edge_weights_fn == "mutual_info":
            edge_weights_fn = mutual_info_score
        elif edge_weights_fn == "adjusted_mutual_info":
            edge_weights_fn = adjusted_mutual_info_score
        elif edge_weights_fn == "normalized_mutual_info":
            edge_weights_fn = normalized_mutual_info_score
        elif not callable(edge_weights_fn):
            raise ValueError(
                f"edge_weights_fn should either be 'mutual_info', 'adjusted_mutual_info', "
                f"'normalized_mutual_info'or a function of form fun(array, array). Got: f{edge_weights_fn}"
            )

        # Step 1: Compute edge weights for a fully connected graph.
        n_vars = len(data.columns)
        pbar = combinations(data.columns, 2)
        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(pbar, total=(n_vars * (n_vars - 1) / 2), desc="Building tree")

        def _conditional_edge_weights_fn(u, v):
            """
            Computes the conditional edge weight of variable index u and v conditioned on class_node
            """
            cond_marginal = data.loc[:, class_node].value_counts() / data.shape[0]
            cond_edge_weight = 0
            for index, marg_prob in cond_marginal.iteritems():
                df_cond_subset = data[data.loc[:, class_node] == index]
                cond_edge_weight += marg_prob * edge_weights_fn(
                    df_cond_subset.loc[:, u], df_cond_subset.loc[:, v]
                )
            return cond_edge_weight

        vals = Parallel(n_jobs=1, prefer="threads")(
            delayed(_conditional_edge_weights_fn)(u, v) for u, v in pbar
        )
        weights = np.zeros((n_vars, n_vars))
        indices = np.triu_indices(n_vars, k=1)
        weights[indices] = vals
        weights.T[indices] = vals

        return weights

    @staticmethod
    def _create_tree_and_dag(weights, columns, root_node):
        """
        Helper function to Chow-Liu algorithm for estimating tree structure from given data. Refer to
        pgmpy.estimators.TreeSearch for more details. This function returns the DAG based on the edge weights matrix.

        Parameters
        ----------
        weights: numpy 2D array, shape = (n_columns, n_columns)
            symmetric matrix where each element represents an edge weight.

        columns: list or array
            Names of the columns (& rows) of the weights matrix.

        root_node: str, int, or any hashable python object.
            The root node of the tree structure.

        Returns
        -------
        model: pgmpy.base.DAG
            The estimated model structure.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.estimators import TreeSearch
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 5)),
        ...                       columns=['A', 'B', 'C', 'D', 'E'])
        >>> est = TreeSearch(values, root_node='B')
        >>> model = est.estimate(estimator_type='chow-liu')
        """
        # Step 2: Compute the maximum spanning tree using the weights.
        T = nx.maximum_spanning_tree(
            nx.from_pandas_adjacency(
                pd.DataFrame(weights, index=columns, columns=columns),
                create_using=nx.Graph,
            )
        )

        # Step 3: Create DAG by directing edges away from root node and return
        D = nx.bfs_tree(T, root_node)
        return DAG(D)



from .metrics import correlation_score, log_likelihood_score, structure_score
from .bn_inference import BayesianModelProbability


__all__ = [
    "correlation_score",
    "log_likelihood_score",
    "structure_score",
]



import warnings

from pgmpy.models import MarkovNetwork


class MarkovModel(MarkovNetwork):
    def __init__(self, ebunch=None, latents=set()):
        warnings.warn(
            "MarkovModel has been renamed to MarkovNetwork. Please use MarkovNetwork class, MarkovModel will be removed in future.",
            FutureWarning,
        )
        super(MarkovModel, self).__init__(ebunch=ebunch, latents=latents)



#!/usr/bin/env python3

from collections import defaultdict
from itertools import chain

import numpy as np

from pgmpy.factors.discrete import TabularCPD
from pgmpy.models import (
    BayesianNetwork,
    DynamicBayesianNetwork,
    FactorGraph,
    JunctionTree,
    MarkovNetwork,
)


class Inference(object):
    """
    Base class for all inference algorithms.

    Converts BayesianNetwork and MarkovNetwork to a uniform representation so that inference
    algorithms can be applied. Also, it checks if all the associated CPDs / Factors are
    consistent with the model.

    Initialize inference for a model.

    Parameters
    ----------
    model: pgmpy.models.BayesianNetwork or pgmpy.models.MarkovNetwork or pgmpy.models.NoisyOrModel
        model for which to initialize the inference object.

    Examples
    --------
    >>> from pgmpy.inference import Inference
    >>> from pgmpy.models import BayesianNetwork
    >>> from pgmpy.factors.discrete import TabularCPD
    >>> student = BayesianNetwork([('diff', 'grade'), ('intel', 'grade')])
    >>> diff_cpd = TabularCPD('diff', 2, [[0.2], [0.8]])
    >>> intel_cpd = TabularCPD('intel', 2, [[0.3], [0.7]])
    >>> grade_cpd = TabularCPD('grade', 3, [[0.1, 0.1, 0.1, 0.1],
    ...                                     [0.1, 0.1, 0.1, 0.1],
    ...                                     [0.8, 0.8, 0.8, 0.8]],
    ...                        evidence=['diff', 'intel'], evidence_card=[2, 2])
    >>> student.add_cpds(diff_cpd, intel_cpd, grade_cpd)
    >>> model = Inference(student)

    >>> from pgmpy.models import MarkovNetwork
    >>> from pgmpy.factors.discrete import DiscreteFactor
    >>> import numpy as np
    >>> student = MarkovNetwork([('Alice', 'Bob'), ('Bob', 'Charles'),
    ...                        ('Charles', 'Debbie'), ('Debbie', 'Alice')])
    >>> factor_a_b = DiscreteFactor(['Alice', 'Bob'], cardinality=[2, 2],
    ...                             values=np.random.rand(4))
    >>> factor_b_c = DiscreteFactor(['Bob', 'Charles'], cardinality=[2, 2],
    ...                             values=np.random.rand(4))
    >>> factor_c_d = DiscreteFactor(['Charles', 'Debbie'], cardinality=[2, 2],
    ...                             values=np.random.rand(4))
    >>> factor_d_a = DiscreteFactor(['Debbie', 'Alice'], cardinality=[2, 2],
    ...                             values=np.random.rand(4))
    >>> student.add_factors(factor_a_b, factor_b_c, factor_c_d, factor_d_a)
    >>> model = Inference(student)
    """

    def __init__(self, model):
        self.model = model
        model.check_model()

        if isinstance(self.model, JunctionTree):
            self.variables = set(chain(*self.model.nodes()))
        else:
            self.variables = self.model.nodes()

    def _initialize_structures(self):
        """
        Initializes all the data structures which will
        later be used by the inference algorithms.
        """
        if isinstance(self.model, JunctionTree):
            self.variables = set(chain(*self.model.nodes()))
        else:
            self.variables = self.model.nodes()

        self.cardinality = {}
        self.factors = defaultdict(list)

        if isinstance(self.model, BayesianNetwork):
            self.state_names_map = {}
            for node in self.model.nodes():
                cpd = self.model.get_cpds(node)
                if isinstance(cpd, TabularCPD):
                    self.cardinality[node] = cpd.variable_card
                    cpd = cpd.to_factor()
                for var in cpd.scope():
                    self.factors[var].append(cpd)
                self.state_names_map.update(cpd.no_to_name)

        elif isinstance(self.model, (MarkovNetwork, FactorGraph, JunctionTree)):
            self.cardinality = self.model.get_cardinality()

            for factor in self.model.get_factors():
                for var in factor.variables:
                    self.factors[var].append(factor)

        elif isinstance(self.model, DynamicBayesianNetwork):
            self.start_bayesian_model = BayesianNetwork(self.model.get_intra_edges(0))
            self.start_bayesian_model.add_cpds(*self.model.get_cpds(time_slice=0))
            cpd_inter = [
                self.model.get_cpds(node) for node in self.model.get_interface_nodes(1)
            ]
            self.interface_nodes = self.model.get_interface_nodes(0)
            self.one_and_half_model = BayesianNetwork(
                self.model.get_inter_edges() + self.model.get_intra_edges(1)
            )
            self.one_and_half_model.add_cpds(
                *(self.model.get_cpds(time_slice=1) + cpd_inter)
            )

    def _prune_bayesian_model(self, variables, evidence):
        """
        Prunes unnecessary nodes from the model to optimize the computation.

        Parameters
        ----------
        variables: list
            The variables on which the query is done i.e. the variables whose
            values we are interested in.

        evidence: dict (default: None)
            The variables whose values we know. The values can be specified as
            {variable: state}.

        Returns
        -------
        Pruned model: pgmpy.models.BayesianNetwork
            The pruned model.

        Examples
        --------
        >>>
        >>>

        References
        ----------
        [1] Baker, M., & Boult, T. E. (2013). Pruning Bayesian networks for efficient computation. arXiv preprint arXiv:1304.1112.
        """
        evidence = {} if evidence is None else evidence
        variables = list(self.model.nodes()) if len(variables) == 0 else list(variables)

        # Step 1: Remove all the variables that are d-separated from `variables` when conditioned
        #         on `evidence`
        d_connected = self.model.active_trail_nodes(
            variables=variables, observed=list(evidence.keys()), include_latents=True
        )
        d_connected = set.union(*d_connected.values()).union(evidence.keys())
        bn = self.model.subgraph(d_connected)
        evidence = {var: state for var, state in evidence.items() if var in d_connected}

        # Step 2: Reduce the model to ancestral graph of [`variables` + `evidence`]
        bn = bn.get_ancestral_graph(list(variables) + list(evidence.keys()))

        # Step 3: Since all the CPDs are lost, add them back. Also marginalize them if some
        #         of the variables in scope aren't in the network anymore.
        cpds = []
        for var in bn.nodes():
            cpd = self.model.get_cpds(var)
            if set(cpd.scope()) < set(bn.nodes()):
                cpds.append(cpd)
            else:
                cpds.append(
                    cpd.marginalize(set(cpd.scope()) - set(bn.nodes()), inplace=False)
                )

        bn.add_cpds(*cpds)

        return bn, evidence

    def _virtual_evidence(self, virtual_evidence):
        """
        Modifies the model to incorporate virtual evidence. For each virtual evidence
        variable a binary variable is added as the child of the evidence variable to
        the model. The state 0 probabilities of the child is the evidence.

        Parameters
        ----------
        virtual_evidence: dict
            A dict of TabularCPD instances specifying the virtual evidence for each
            of the evidence variables.

        Returns
        -------
        None

        References
        ----------
        [1] Mrad, Ali Ben, et al. "Uncertain evidence in Bayesian networks: Presentation and comparison on a simple example." International Conference on Information Processing and Management of Uncertainty in Knowledge-Based Systems. Springer, Berlin, Heidelberg, 2012.
        """
        for cpd in virtual_evidence:
            var = cpd.variables[0]
            if var not in self.model.nodes():
                raise ValueError(
                    "Evidence provided for variable which is not in the model"
                )
            elif len(cpd.variables) > 1:
                raise ValueError(
                    "Virtual evidence should be defined on individual variables. Maybe you are looking for soft evidence."
                )

            elif self.model.get_cardinality(var) != cpd.get_cardinality([var])[var]:
                raise ValueError(
                    "The number of states/cardinality for the evidence should be same as the number of states/cardinality of the variable in the model"
                )

        bn = self.model.copy()
        for cpd in virtual_evidence:
            var = cpd.variables[0]
            new_var = "__" + var
            bn.add_edge(var, new_var)
            values = np.vstack((cpd.values, 1 - cpd.values))
            new_cpd = TabularCPD(
                variable=new_var,
                variable_card=2,
                values=values,
                evidence=[var],
                evidence_card=[self.model.get_cardinality(var)],
                state_names={new_var: [0, 1], var: cpd.state_names[var]},
            )
            bn.add_cpds(new_cpd)

        self.__init__(bn)



# -*- coding: utf-8 -*-

import numbers
from itertools import chain
from warnings import warn

import numpy as np
from joblib import Parallel, delayed

from pgmpy.estimators import ParameterEstimator
from pgmpy.factors.discrete import TabularCPD
from pgmpy.models import BayesianNetwork


class BayesianEstimator(ParameterEstimator):
    def __init__(self, model, data, **kwargs):
        """
        Class used to compute parameters for a model using Bayesian Parameter Estimation.
        See `MaximumLikelihoodEstimator` for constructor parameters.
        """
        if not isinstance(model, BayesianNetwork):
            raise NotImplementedError(
                "Bayesian Parameter Estimation is only implemented for BayesianNetwork"
            )
        elif len(model.latents) != 0:
            raise ValueError(
                f"Bayesian Parameter Estimation works only on models with all observed variables. Found latent variables: {model.latents}"
            )

        super(BayesianEstimator, self).__init__(model, data, **kwargs)

    def get_parameters(
        self,
        prior_type="BDeu",
        equivalent_sample_size=5,
        pseudo_counts=None,
        n_jobs=-1,
        weighted=False,
    ):
        """
        Method to estimate the model parameters (CPDs).

        Parameters
        ----------
        prior_type: 'dirichlet', 'BDeu', or 'K2'
            string indicting which type of prior to use for the model parameters.
            - If 'prior_type' is 'dirichlet', the following must be provided:
                'pseudo_counts' = dirichlet hyperparameters; a single number or a dict containing, for each
                 variable, a 2-D array of the shape (node_card, product of parents_card) with a "virtual"
                 count for each variable state in the CPD, that is added to the state counts.
                 (lexicographic ordering of states assumed)
            - If 'prior_type' is 'BDeu', then an 'equivalent_sample_size'
                must be specified instead of 'pseudo_counts'. This is equivalent to
                'prior_type=dirichlet' and using uniform 'pseudo_counts' of
                `equivalent_sample_size/(node_cardinality*np.prod(parents_cardinalities))` for each node.
                'equivalent_sample_size' can either be a numerical value or a dict that specifies
                the size for each variable separately.
            - A prior_type of 'K2' is a shorthand for 'dirichlet' + setting every pseudo_count to 1,
                regardless of the cardinality of the variable.

        weighted: bool
            If weighted=True, the data must contain a `_weight` column specifying the
            weight of each datapoint (row). If False, assigns an equal weight to each
            datapoint.

        Returns
        -------
        parameters: list
            List of TabularCPDs, one for each variable of the model

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.estimators import BayesianEstimator
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 4)),
        ...                       columns=['A', 'B', 'C', 'D'])
        >>> model = BayesianNetwork([('A', 'B'), ('C', 'B'), ('C', 'D')])
        >>> estimator = BayesianEstimator(model, values)
        >>> estimator.get_parameters(prior_type='BDeu', equivalent_sample_size=5)
        [<TabularCPD representing P(C:2) at 0x7f7b534251d0>,
        <TabularCPD representing P(B:2 | C:2, A:2) at 0x7f7b4dfd4da0>,
        <TabularCPD representing P(A:2) at 0x7f7b4dfd4fd0>,
        <TabularCPD representing P(D:2 | C:2) at 0x7f7b4df822b0>]
        """

        def _get_node_param(node):
            _equivalent_sample_size = (
                equivalent_sample_size[node]
                if isinstance(equivalent_sample_size, dict)
                else equivalent_sample_size
            )
            if isinstance(pseudo_counts, numbers.Real):
                _pseudo_counts = pseudo_counts
            else:
                _pseudo_counts = pseudo_counts[node] if pseudo_counts else None

            cpd = self.estimate_cpd(
                node,
                prior_type=prior_type,
                equivalent_sample_size=_equivalent_sample_size,
                pseudo_counts=_pseudo_counts,
                weighted=weighted,
            )
            return cpd

        parameters = Parallel(n_jobs=n_jobs, prefer="threads")(
            delayed(_get_node_param)(node) for node in self.model.nodes()
        )

        return parameters

    def estimate_cpd(
        self,
        node,
        prior_type="BDeu",
        pseudo_counts=[],
        equivalent_sample_size=5,
        weighted=False,
    ):
        """
        Method to estimate the CPD for a given variable.

        Parameters
        ----------
        node: int, string (any hashable python object)
            The name of the variable for which the CPD is to be estimated.

        prior_type: 'dirichlet', 'BDeu', 'K2',
            string indicting which type of prior to use for the model parameters.
            - If 'prior_type' is 'dirichlet', the following must be provided:
                'pseudo_counts' = dirichlet hyperparameters; a single number or 2-D array
                 of shape (node_card, product of parents_card) with a "virtual" count for
                 each variable state in the CPD. The virtual counts are added to the
                 actual state counts found in the data. (if a list is provided, a
                 lexicographic ordering of states is assumed)
            - If 'prior_type' is 'BDeu', then an 'equivalent_sample_size'
                 must be specified instead of 'pseudo_counts'. This is equivalent to
                 'prior_type=dirichlet' and using uniform 'pseudo_counts' of
                 `equivalent_sample_size/(node_cardinality*np.prod(parents_cardinalities))`.
            - A prior_type of 'K2' is a shorthand for 'dirichlet' + setting every
              pseudo_count to 1, regardless of the cardinality of the variable.

        weighted: bool
            If weighted=True, the data must contain a `_weight` column specifying the
            weight of each datapoint (row). If False, assigns an equal weight to each
            datapoint.

        Returns
        -------
        CPD: TabularCPD
            The estimated CPD for `node`.

        Examples
        --------
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.estimators import BayesianEstimator
        >>> data = pd.DataFrame(data={'A': [0, 0, 1], 'B': [0, 1, 0], 'C': [1, 1, 0]})
        >>> model = BayesianNetwork([('A', 'C'), ('B', 'C')])
        >>> estimator = BayesianEstimator(model, data)
        >>> cpd_C = estimator.estimate_cpd('C', prior_type="dirichlet",
        ...                                pseudo_counts=[[1, 1, 1, 1],
        ...                                               [2, 2, 2, 2]])
        >>> print(cpd_C)
        ╒══════╤══════╤══════╤══════╤════════════════════╕
        │ A    │ A(0) │ A(0) │ A(1) │ A(1)               │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ B    │ B(0) │ B(1) │ B(0) │ B(1)               │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ C(0) │ 0.25 │ 0.25 │ 0.5  │ 0.3333333333333333 │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ C(1) │ 0.75 │ 0.75 │ 0.5  │ 0.6666666666666666 │
        ╘══════╧══════╧══════╧══════╧════════════════════╛

        """
        node_cardinality = len(self.state_names[node])
        parents = sorted(self.model.get_parents(node))
        parents_cardinalities = [len(self.state_names[parent]) for parent in parents]
        cpd_shape = (node_cardinality, np.prod(parents_cardinalities, dtype=int))

        prior_type = prior_type.lower()

        # Throw a warning if pseudo_count is specified without prior_type=dirichlet
        #     cast to np.array first to use the array.size attribute, which returns 0 also for [[],[]]
        #     (where len([[],[]]) evaluates to 2)
        if (
            pseudo_counts is not None
            and np.array(pseudo_counts).size > 0
            and (prior_type != "dirichlet")
        ):
            warn(
                f"pseudo count specified with {prior_type} prior. It will be ignored, use dirichlet prior for specifying pseudo_counts"
            )

        if prior_type == "k2":
            pseudo_counts = np.ones(cpd_shape, dtype=int)
        elif prior_type == "bdeu":
            alpha = float(equivalent_sample_size) / (
                node_cardinality * np.prod(parents_cardinalities)
            )
            pseudo_counts = np.ones(cpd_shape, dtype=float) * alpha
        elif prior_type == "dirichlet":
            if isinstance(pseudo_counts, numbers.Real):
                pseudo_counts = np.ones(cpd_shape, dtype=int) * pseudo_counts

            else:
                pseudo_counts = np.array(pseudo_counts)
                if pseudo_counts.shape != cpd_shape:
                    raise ValueError(
                        f"The shape of pseudo_counts for the node: {node} must be of shape: {str(cpd_shape)}"
                    )
        else:
            raise ValueError("'prior_type' not specified")

        state_counts = self.state_counts(node, weighted=weighted)
        bayesian_counts = state_counts + pseudo_counts

        cpd = TabularCPD(
            node,
            node_cardinality,
            np.array(bayesian_counts),
            evidence=parents,
            evidence_card=parents_cardinalities,
            state_names={var: self.state_names[var] for var in chain([node], parents)},
        )
        cpd.normalize()
        return cpd



#!/usr/bin/env python3

import unittest

import networkx as nx
import numpy as np

import pgmpy.tests.help_functions as hf
from pgmpy.base import DAG, PDAG


class TestDAGCreation(unittest.TestCase):
    def setUp(self):
        self.graph = DAG()

    def test_class_init_without_data(self):
        self.assertIsInstance(self.graph, DAG)

    def test_class_init_with_data_string(self):
        self.graph = DAG([("a", "b"), ("b", "c")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["b", "c"]]
        )
        self.assertEqual(self.graph.latents, set())

        self.graph = DAG([("a", "b"), ("b", "c")], latents=["b"])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["b", "c"]]
        )
        self.assertEqual(self.graph.latents, set(["b"]))

    def test_add_node_string(self):
        self.graph = DAG()
        self.graph.add_node("a")
        self.assertListEqual(list(self.graph.nodes()), ["a"])
        self.assertEqual(self.graph.latents, set())

        self.graph = DAG()
        self.graph.add_node("a", latent=True)
        self.assertListEqual(list(self.graph.nodes()), ["a"])
        self.assertEqual(self.graph.latents, set(["a"]))

    def test_add_node_nonstring(self):
        self.graph = DAG()
        self.graph.add_node(1)

        self.graph = DAG()
        self.graph.add_node(1, latent=True)

    def test_add_nodes_from_string(self):
        self.graph = DAG()
        self.graph.add_nodes_from(["a", "b", "c", "d"])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d"])
        self.assertEqual(self.graph.latents, set())

        self.graph = DAG()
        self.graph.add_nodes_from(["a", "b", "c", "d"], latent=True)
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d"])
        self.assertEqual(self.graph.latents, set(["a", "b", "c", "d"]))

        self.graph = DAG()
        self.graph.add_nodes_from(
            ["a", "b", "c", "d"], latent=[True, False, True, False]
        )
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d"])
        self.assertEqual(self.graph.latents, set(["a", "c"]))

    def test_add_nodes_from_non_string(self):
        self.graph = DAG()
        self.graph.add_nodes_from([1, 2, 3, 4])

        self.graph = DAG()
        self.graph.add_nodes_from([1, 2, 3, 4], latent=True)

        self.graph = DAG()
        self.graph.add_nodes_from([1, 2, 3, 4], latent=[True, False, False, False])

    def test_add_node_weight(self):
        self.graph = DAG()
        self.graph.add_node("weighted_a", weight=0.3)
        self.assertEqual(self.graph.nodes["weighted_a"]["weight"], 0.3)
        self.assertEqual(self.graph.latents, set())

        self.graph = DAG()
        self.graph.add_node("weighted_a", weight=0.3, latent=True)
        self.assertEqual(self.graph.nodes["weighted_a"]["weight"], 0.3)
        self.assertEqual(self.graph.latents, set(["weighted_a"]))

    def test_add_nodes_from_weight(self):
        self.graph = DAG()
        self.graph.add_nodes_from(["weighted_b", "weighted_c"], weights=[0.5, 0.6])
        self.assertEqual(self.graph.nodes["weighted_b"]["weight"], 0.5)
        self.assertEqual(self.graph.nodes["weighted_c"]["weight"], 0.6)

        self.graph = DAG()
        self.graph.add_nodes_from(
            ["weighted_b", "weighted_c"], weights=[0.5, 0.6], latent=True
        )
        self.assertEqual(self.graph.nodes["weighted_b"]["weight"], 0.5)
        self.assertEqual(self.graph.nodes["weighted_c"]["weight"], 0.6)
        self.assertEqual(self.graph.latents, set(["weighted_b", "weighted_c"]))

        self.graph = DAG()
        self.graph.add_nodes_from(
            ["weighted_b", "weighted_c"], weights=[0.5, 0.6], latent=[True, False]
        )
        self.assertEqual(self.graph.nodes["weighted_b"]["weight"], 0.5)
        self.assertEqual(self.graph.nodes["weighted_c"]["weight"], 0.6)
        self.assertEqual(self.graph.latents, set(["weighted_b"]))

        self.graph.add_nodes_from(["e", "f"])
        self.assertEqual(self.graph.nodes["e"]["weight"], None)
        self.assertEqual(self.graph.nodes["f"]["weight"], None)

    def test_add_edge_string(self):
        self.graph.add_edge("d", "e")
        self.assertListEqual(sorted(self.graph.nodes()), ["d", "e"])
        self.assertListEqual(list(self.graph.edges()), [("d", "e")])
        self.graph.add_nodes_from(["a", "b", "c"])
        self.graph.add_edge("a", "b")
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["d", "e"]]
        )

    def test_add_edge_nonstring(self):
        self.graph.add_edge(1, 2)

    def test_add_edges_from_string(self):
        self.graph.add_edges_from([("a", "b"), ("b", "c")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()), [["a", "b"], ["b", "c"]]
        )
        self.graph.add_nodes_from(["d", "e", "f"])
        self.graph.add_edges_from([("d", "e"), ("e", "f")])
        self.assertListEqual(sorted(self.graph.nodes()), ["a", "b", "c", "d", "e", "f"])
        self.assertListEqual(
            hf.recursive_sorted(self.graph.edges()),
            hf.recursive_sorted([("a", "b"), ("b", "c"), ("d", "e"), ("e", "f")]),
        )

    def test_add_edges_from_nonstring(self):
        self.graph.add_edges_from([(1, 2), (2, 3)])

    def test_add_edge_weight(self):
        self.graph.add_edge("a", "b", weight=0.3)
        if nx.__version__.startswith("1"):
            self.assertEqual(self.graph.edge["a"]["b"]["weight"], 0.3)
        else:
            self.assertEqual(self.graph.adj["a"]["b"]["weight"], 0.3)

    def test_add_edges_from_weight(self):
        self.graph.add_edges_from([("b", "c"), ("c", "d")], weights=[0.5, 0.6])
        if nx.__version__.startswith("1"):
            self.assertEqual(self.graph.edge["b"]["c"]["weight"], 0.5)
            self.assertEqual(self.graph.edge["c"]["d"]["weight"], 0.6)

            self.graph.add_edges_from([("e", "f")])
            self.assertEqual(self.graph.edge["e"]["f"]["weight"], None)
        else:
            self.assertEqual(self.graph.adj["b"]["c"]["weight"], 0.5)
            self.assertEqual(self.graph.adj["c"]["d"]["weight"], 0.6)

            self.graph.add_edges_from([("e", "f")])
            self.assertEqual(self.graph.adj["e"]["f"]["weight"], None)

    def test_update_node_parents_bm_constructor(self):
        self.graph = DAG([("a", "b"), ("b", "c")])
        self.assertListEqual(list(self.graph.predecessors("a")), [])
        self.assertListEqual(list(self.graph.predecessors("b")), ["a"])
        self.assertListEqual(list(self.graph.predecessors("c")), ["b"])

    def test_update_node_parents(self):
        self.graph.add_nodes_from(["a", "b", "c"])
        self.graph.add_edges_from([("a", "b"), ("b", "c")])
        self.assertListEqual(list(self.graph.predecessors("a")), [])
        self.assertListEqual(list(self.graph.predecessors("b")), ["a"])
        self.assertListEqual(list(self.graph.predecessors("c")), ["b"])

    def test_get_leaves(self):
        self.graph.add_edges_from(
            [("A", "B"), ("B", "C"), ("B", "D"), ("D", "E"), ("D", "F"), ("A", "G")]
        )
        self.assertEqual(sorted(self.graph.get_leaves()), sorted(["C", "G", "E", "F"]))

    def test_get_roots(self):
        self.graph.add_edges_from(
            [("A", "B"), ("B", "C"), ("B", "D"), ("D", "E"), ("D", "F"), ("A", "G")]
        )
        self.assertEqual(["A"], self.graph.get_roots())
        self.graph.add_edge("H", "G")
        self.assertEqual(sorted(["A", "H"]), sorted(self.graph.get_roots()))

    def test_init_with_cycle(self):
        self.assertRaises(ValueError, DAG, [("a", "a")])
        self.assertRaises(ValueError, DAG, [("a", "b"), ("b", "a")])
        self.assertRaises(ValueError, DAG, [("a", "b"), ("b", "c"), ("c", "a")])

    def test_get_ancestral_graph(self):
        dag = DAG([("A", "C"), ("B", "C"), ("D", "A"), ("D", "B")])
        anc_dag = dag.get_ancestral_graph(["A", "B"])
        self.assertEqual(set(anc_dag.edges()), set([("D", "A"), ("D", "B")]))
        self.assertRaises(ValueError, dag.get_ancestral_graph, ["A", "gibber"])

    def test_minimal_dseparator(self):
        # Without latent variables

        dag1 = DAG([("A", "B"), ("B", "C")])
        self.assertEqual(dag1.minimal_dseparator(start="A", end="C"), {"B"})

        dag2 = DAG([("A", "B"), ("B", "C"), ("C", "D"), ("A", "E"), ("E", "D")])
        self.assertEqual(dag2.minimal_dseparator(start="A", end="D"), {"C", "E"})

        dag3 = DAG(
            [("B", "A"), ("B", "C"), ("A", "D"), ("D", "C"), ("A", "E"), ("C", "E")]
        )
        self.assertEqual(dag3.minimal_dseparator(start="A", end="C"), {"B", "D"})

        # With latent variables

        dag_lat1 = DAG([("A", "B"), ("B", "C")], latents={"B"})
        self.assertIsNone(dag_lat1.minimal_dseparator(start="A", end="C"))

        dag_lat2 = DAG([("A", "D"), ("D", "B"), ("B", "C")], latents={"B"})
        self.assertEqual(dag_lat2.minimal_dseparator(start="A", end="C"), {"D"})

        dag_lat3 = DAG([("A", "B"), ("B", "D"), ("D", "C")], latents={"B"})
        self.assertEqual(dag_lat3.minimal_dseparator(start="A", end="C"), {"D"})

        dag_lat4 = DAG([("A", "B"), ("B", "C"), ("A", "D"), ("D", "C")], latents={"D"})
        self.assertIsNone(dag_lat4.minimal_dseparator(start="A", end="C"))

        dag_lat5 = DAG(
            [("A", "B"), ("B", "C"), ("A", "D"), ("D", "E"), ("E", "C")], latents={"E"}
        )
        self.assertEqual(dag_lat5.minimal_dseparator(start="A", end="C"), {"B", "D"})

    def test_to_daft(self):
        dag = DAG([("A", "C"), ("B", "C"), ("D", "A"), ("D", "B")])
        dag.to_daft(node_pos="circular")

        self.assertRaises(ValueError, dag.to_daft, node_pos="abcd")
        self.assertRaises(ValueError, dag.to_daft, node_pos={"A": (0, 2)})
        self.assertRaises(ValueError, dag.to_daft, node_pos=["random"])

        for layout in [
            "circular",
            "kamada_kawai",
            "planar",
            "random",
            "shell",
            "spring",
            #            "spectral", # TODO: Fails for latest networkx
            "spiral",
        ]:
            dag.to_daft(node_pos=layout)
            dag.to_daft(node_pos=layout, pgm_params={"observed_style": "inner"})
            dag.to_daft(
                node_pos=layout,
                edge_params={("A", "C"): {"label": 2}},
                node_params={"A": {"shape": "rectangle"}},
            )

    def test_random_dag(self):
        for i in range(100):
            n_nodes = np.random.randint(low=2, high=100)
            edge_prob = np.random.uniform()
            dag = DAG.get_random(n_nodes=n_nodes, edge_prob=edge_prob)
            self.assertEqual(len(dag.nodes()), n_nodes)
            self.assertTrue(nx.is_directed_acyclic_graph(dag))
            self.assertTrue(len(dag.latents) == 0)

        dag_latents = DAG.get_random(n_nodes=n_nodes, edge_prob=0.5, latents=True)

    def tearDown(self):
        del self.graph


class TestDAGMoralization(unittest.TestCase):
    def setUp(self):
        self.graph = DAG()
        self.graph.add_edges_from([("diff", "grade"), ("intel", "grade")])

    def test_get_parents(self):
        self.assertListEqual(sorted(self.graph.get_parents("grade")), ["diff", "intel"])

    def test_moralize(self):
        moral_graph = self.graph.moralize()
        self.assertListEqual(
            hf.recursive_sorted(moral_graph.edges()),
            [["diff", "grade"], ["diff", "intel"], ["grade", "intel"]],
        )

    def test_moralize_disconnected(self):
        graph_copy = self.graph.copy()
        graph_copy.add_node("disconnected")
        moral_graph = graph_copy.moralize()
        self.assertListEqual(
            hf.recursive_sorted(moral_graph.edges()),
            [["diff", "grade"], ["diff", "intel"], ["grade", "intel"]],
        )
        self.assertEqual(
            sorted(moral_graph.nodes()), ["diff", "disconnected", "grade", "intel"]
        )

    def test_get_children(self):
        self.assertListEqual(sorted(self.graph.get_children("diff")), ["grade"])

    def tearDown(self):
        del self.graph


class TestDoOperator(unittest.TestCase):
    def setUp(self):
        self.g1 = DAG([("X", "A"), ("A", "Y"), ("A", "B")])
        self.g2 = DAG([("X", "A"), ("A", "Y"), ("A", "B"), ("Z", "Y")])

    def test_do(self):
        dag_do_x = self.g1.do("A")
        self.assertEqual(set(dag_do_x.nodes()), set(self.g1.nodes()))
        self.assertEqual(sorted(list(dag_do_x.edges())), [("A", "B"), ("A", "Y")])

        dag_do_x = self.g2.do(["A", "Y"])
        self.assertEqual(set(dag_do_x.nodes()), set(self.g2.nodes()))
        self.assertEqual(sorted(list(dag_do_x.edges())), [("A", "B")])


class TestPDAG(unittest.TestCase):
    def setUp(self):
        self.pdag_mix = PDAG(
            directed_ebunch=[("A", "C"), ("D", "C")],
            undirected_ebunch=[("B", "A"), ("B", "D")],
        )
        self.pdag_dir = PDAG(
            directed_ebunch=[("A", "B"), ("D", "B"), ("A", "C"), ("D", "C")]
        )
        self.pdag_undir = PDAG(
            undirected_ebunch=[("A", "C"), ("D", "C"), ("B", "A"), ("B", "D")]
        )
        self.pdag_latent = PDAG(
            directed_ebunch=[("A", "C"), ("D", "C")],
            undirected_ebunch=[("B", "A"), ("B", "D")],
            latents=["A", "D"],
        )

    def test_init_normal(self):
        # Mix directed and undirected
        directed_edges = [("A", "C"), ("D", "C")]
        undirected_edges = [("B", "A"), ("B", "D")]
        pdag = PDAG(directed_ebunch=directed_edges, undirected_ebunch=undirected_edges)
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))

        directed_edges = [("A", "C"), ("D", "C")]
        undirected_edges = [("B", "A"), ("B", "D")]
        pdag = PDAG(
            directed_ebunch=directed_edges,
            undirected_ebunch=undirected_edges,
            latents=["A", "C"],
        )
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))
        self.assertEqual(pdag.latents, set(["A", "C"]))

        # Only undirected
        undirected_edges = [("A", "C"), ("D", "C"), ("B", "A"), ("B", "D")]
        pdag = PDAG(undirected_ebunch=undirected_edges)
        expected_edges = {
            ("A", "C"),
            ("C", "A"),
            ("D", "C"),
            ("C", "D"),
            ("B", "A"),
            ("A", "B"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set())
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))

        undirected_edges = [("A", "C"), ("D", "C"), ("B", "A"), ("B", "D")]
        pdag = PDAG(undirected_ebunch=undirected_edges, latents=["A", "D"])
        expected_edges = {
            ("A", "C"),
            ("C", "A"),
            ("D", "C"),
            ("C", "D"),
            ("B", "A"),
            ("A", "B"),
            ("B", "D"),
            ("D", "B"),
        }
        self.assertEqual(set(pdag.edges()), expected_edges)
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set())
        self.assertEqual(pdag.undirected_edges, set(undirected_edges))
        self.assertEqual(pdag.latents, set(["A", "D"]))

        # Only directed
        directed_edges = [("A", "B"), ("D", "B"), ("A", "C"), ("D", "C")]
        pdag = PDAG(directed_ebunch=directed_edges)
        self.assertEqual(set(pdag.edges()), set(directed_edges))
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set())

        directed_edges = [("A", "B"), ("D", "B"), ("A", "C"), ("D", "C")]
        pdag = PDAG(directed_ebunch=directed_edges, latents=["D"])
        self.assertEqual(set(pdag.edges()), set(directed_edges))
        self.assertEqual(set(pdag.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag.directed_edges, set(directed_edges))
        self.assertEqual(pdag.undirected_edges, set())
        self.assertEqual(pdag.latents, set(["D"]))

        # TODO: Fix the cycle issue.
        # Test cycle
        # directed_edges = [('A', 'C')]
        # undirected_edges = [('A', 'B'), ('B', 'D'), ('D', 'C')]
        # self.assertRaises(ValueError, PDAG, directed_ebunch=directed_edges, undirected_ebunch=undirected_edges)

    def test_copy(self):
        pdag_copy = self.pdag_mix.copy()
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        expected_dir = [("A", "C"), ("D", "C")]
        expected_undir = [("B", "A"), ("B", "D")]
        self.assertEqual(set(pdag_copy.edges()), expected_edges)
        self.assertEqual(set(pdag_copy.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag_copy.directed_edges, set([("A", "C"), ("D", "C")]))
        self.assertEqual(pdag_copy.undirected_edges, set([("B", "A"), ("B", "D")]))
        self.assertEqual(pdag_copy.latents, set())

        pdag_copy = self.pdag_latent.copy()
        expected_edges = {
            ("A", "C"),
            ("D", "C"),
            ("A", "B"),
            ("B", "A"),
            ("B", "D"),
            ("D", "B"),
        }
        expected_dir = [("A", "C"), ("D", "C")]
        expected_undir = [("B", "A"), ("B", "D")]
        self.assertEqual(set(pdag_copy.edges()), expected_edges)
        self.assertEqual(set(pdag_copy.nodes()), {"A", "B", "C", "D"})
        self.assertEqual(pdag_copy.directed_edges, set([("A", "C"), ("D", "C")]))
        self.assertEqual(pdag_copy.undirected_edges, set([("B", "A"), ("B", "D")]))
        self.assertEqual(pdag_copy.latents, set(["A", "D"]))

    def test_pdag_to_dag(self):
        # PDAG no: 1  Possibility of creating a v-structure
        pdag = PDAG(
            directed_ebunch=[("A", "B"), ("C", "B")],
            undirected_ebunch=[("C", "D"), ("D", "A")],
        )
        dag = pdag.to_dag()
        self.assertTrue(("A", "B") in dag.edges())
        self.assertTrue(("C", "B") in dag.edges())
        self.assertFalse((("A", "D") in dag.edges()) and (("C", "D") in dag.edges()))
        self.assertTrue(len(dag.edges()) == 4)

        # With latents
        pdag = PDAG(
            directed_ebunch=[("A", "B"), ("C", "B")],
            undirected_ebunch=[("C", "D"), ("D", "A")],
            latents=["A"],
        )
        dag = pdag.to_dag()
        self.assertTrue(("A", "B") in dag.edges())
        self.assertTrue(("C", "B") in dag.edges())
        self.assertFalse((("A", "D") in dag.edges()) and (("C", "D") in dag.edges()))
        self.assertEqual(dag.latents, set(["A"]))
        self.assertTrue(len(dag.edges()) == 4)

        # PDAG no: 2  No possibility of creation of v-structure.
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")], undirected_ebunch=[("A", "D")]
        )
        dag = pdag.to_dag()
        self.assertTrue(("B", "C") in dag.edges())
        self.assertTrue(("A", "C") in dag.edges())
        self.assertTrue((("A", "D") in dag.edges()) or (("D", "A") in dag.edges()))

        # With latents
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")],
            undirected_ebunch=[("A", "D")],
            latents=["A"],
        )
        dag = pdag.to_dag()
        self.assertTrue(("B", "C") in dag.edges())
        self.assertTrue(("A", "C") in dag.edges())
        self.assertTrue((("A", "D") in dag.edges()) or (("D", "A") in dag.edges()))
        self.assertEqual(dag.latents, set(["A"]))

        # PDAG no: 3  Already existing v-structure, possibility to add another
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")], undirected_ebunch=[("C", "D")]
        )
        dag = pdag.to_dag()
        expected_edges = {("B", "C"), ("C", "D"), ("A", "C")}
        self.assertEqual(expected_edges, set(dag.edges()))

        # With latents
        pdag = PDAG(
            directed_ebunch=[("B", "C"), ("A", "C")],
            undirected_ebunch=[("C", "D")],
            latents=["A"],
        )
        dag = pdag.to_dag()
        expected_edges = {("B", "C"), ("C", "D"), ("A", "C")}
        self.assertEqual(expected_edges, set(dag.edges()))
        self.assertEqual(dag.latents, set(["A"]))



#!/usr/bin/env python
from functools import lru_cache

import pandas as pd

from pgmpy.utils.decorators import convert_args_tuple


class BaseEstimator(object):
    def __init__(self, data=None, state_names=None, complete_samples_only=True):
        """
        Base class for estimators in pgmpy; `ParameterEstimator`,
        `StructureEstimator` and `StructureScore` derive from this class.

        Parameters
        ----------
        data: pandas DataFrame object
            object where each column represents one variable.
            (If some values in the data are missing the data cells should be set to `numpy.NaN`.
            Note that pandas converts each column containing `numpy.NaN`s to dtype `float`.)

        state_names: dict (optional)
            A dict indicating, for each variable, the discrete set of states (or values)
            that the variable can take. If unspecified, the observed values in the data set
            are taken to be the only possible states.

        complete_samples_only: bool (optional, default `True`)
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.Nan` somewhere are ignored. If `False` then, for each variable,
            every row where neither the variable nor its parents are `np.NaN` is used.
            This sets the behavior of the `state_count`-method.
        """

        self.data = data
        # data can be None in the case when learning structure from
        # independence conditions. Look into PC.py.
        if self.data is not None:
            self.complete_samples_only = complete_samples_only

            self.variables = list(data.columns.values)

            if not isinstance(state_names, dict):
                self.state_names = {
                    var: self._collect_state_names(var) for var in self.variables
                }
            else:
                self.state_names = dict()
                for var in self.variables:
                    if var in state_names:
                        if not set(self._collect_state_names(var)) <= set(
                            state_names[var]
                        ):
                            raise ValueError(
                                f"Data contains unexpected states for variable: {var}."
                            )
                        self.state_names[var] = state_names[var]
                    else:
                        self.state_names[var] = self._collect_state_names(var)

    def _collect_state_names(self, variable):
        "Return a list of states that the variable takes in the data."
        states = sorted(list(self.data.loc[:, variable].dropna().unique()))
        return states

    @convert_args_tuple
    @lru_cache(maxsize=2048)
    def state_counts(
        self, variable, parents=[], complete_samples_only=None, weighted=False
    ):
        """
        Return counts how often each state of 'variable' occurred in the data.
        If a list of parents is provided, counting is done conditionally
        for each state configuration of the parents.

        Parameters
        ----------
        variable: string
            Name of the variable for which the state count is to be done.

        parents: list
            Optional list of variable parents, if conditional counting is desired.
            Order of parents in list is reflected in the returned DataFrame

        complete_samples_only: bool
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.NaN` somewhere are ignored. If `False` then
            every row where neither the variable nor its parents are `np.NaN` is used.
            Desired default behavior can be passed to the class constructor.

        weighted: bool
            If True, data must have a `_weight` column specifying the weight of the
            datapoint (row). If False, each datapoint has a weight of `1`.

        Returns
        -------
        state_counts: pandas.DataFrame
            Table with state counts for 'variable'

        Examples
        --------
        >>> import pandas as pd
        >>> from pgmpy.estimators import BaseEstimator
        >>> data = pd.DataFrame(data={'A': ['a1', 'a1', 'a2'],
                                      'B': ['b1', 'b2', 'b1'],
                                      'C': ['c1', 'c1', 'c2']})
        >>> estimator = BaseEstimator(data)
        >>> estimator.state_counts('A')
            A
        a1  2
        a2  1
        >>> estimator.state_counts('C', parents=['A', 'B'])
        A  a1      a2
        B  b1  b2  b1  b2
        C
        c1  1   1   0   0
        c2  0   0   1   0
        >>> estimator.state_counts('C', parents=['A'])
        A    a1   a2
        C
        c1  2.0  0.0
        c2  0.0  1.0
        """
        parents = list(parents)

        # default for how to deal with missing data can be set in class constructor
        if complete_samples_only is None:
            complete_samples_only = self.complete_samples_only
        # ignores either any row containing NaN, or only those where the variable or its parents is NaN
        data = (
            self.data.dropna()
            if complete_samples_only
            else self.data.dropna(subset=[variable] + parents)
        )

        if weighted and ("_weight" not in self.data.columns):
            raise ValueError("data must contain a `_weight` column if weighted=True")

        if not parents:
            # count how often each state of 'variable' occurred
            if weighted:
                state_count_data = data.groupby([variable]).sum()["_weight"]
            else:
                state_count_data = data.loc[:, variable].value_counts()

            state_counts = (
                state_count_data.reindex(self.state_names[variable])
                .fillna(0)
                .to_frame()
            )

        else:
            parents_states = [self.state_names[parent] for parent in parents]
            # count how often each state of 'variable' occurred, conditional on parents' states
            if weighted:
                state_count_data = (
                    data.groupby([variable] + parents).sum()["_weight"].unstack(parents)
                )

            else:
                state_count_data = (
                    data.groupby([variable] + parents).size().unstack(parents)
                )

            if not isinstance(state_count_data.columns, pd.MultiIndex):
                state_count_data.columns = pd.MultiIndex.from_arrays(
                    [state_count_data.columns]
                )

            # reindex rows & columns to sort them and to add missing ones
            # missing row    = some state of 'variable' did not occur in data
            # missing column = some state configuration of current 'variable's parents
            #                  did not occur in data
            row_index = self.state_names[variable]
            column_index = pd.MultiIndex.from_product(parents_states, names=parents)
            state_counts = state_count_data.reindex(
                index=row_index, columns=column_index
            ).fillna(0)

        return state_counts


class ParameterEstimator(BaseEstimator):
    def __init__(self, model, data, **kwargs):
        """
        Base class for parameter estimators in pgmpy.

        Parameters
        ----------
        model: pgmpy.models.BayesianNetwork or pgmpy.models.MarkovNetwork or pgmpy.models.NoisyOrModel model
            for which parameter estimation is to be done.

        data: pandas DataFrame object
            dataframe object with column names identical to the variable names of the model.
            (If some values in the data are missing the data cells should be set to `numpy.NaN`.
            Note that pandas converts each column containing `numpy.NaN`s to dtype `float`.)

        state_names: dict (optional)
            A dict indicating, for each variable, the discrete set of states (or values)
            that the variable can take. If unspecified, the observed values in the data set
            are taken to be the only possible states.

        complete_samples_only: bool (optional, default `True`)
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.Nan` somewhere are ignored. If `False` then, for each variable,
            every row where neither the variable nor its parents are `np.NaN` is used.
            This sets the behavior of the `state_count`-method.
        """

        if not (set(model.nodes()) - model.latents) <= set(data.columns.values):
            raise ValueError(
                "variable names of the model must be identical to column names in data"
            )
        self.model = model

        super(ParameterEstimator, self).__init__(data, **kwargs)

    def state_counts(self, variable, weighted=False, **kwargs):
        """
        Return counts how often each state of 'variable' occurred in the data.
        If the variable has parents, counting is done conditionally
        for each state configuration of the parents.

        Parameters
        ----------
        variable: string
            Name of the variable for which the state count is to be done.

        complete_samples_only: bool
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.NaN` somewhere are ignored. If `False` then
            every row where neither the variable nor its parents are `np.NaN` is used.
            Desired default behavior can be passed to the class constructor.

        Returns
        -------
        state_counts: pandas.DataFrame
            Table with state counts for 'variable'

        Examples
        --------
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.estimators import ParameterEstimator
        >>> model = BayesianNetwork([('A', 'C'), ('B', 'C')])
        >>> data = pd.DataFrame(data={'A': ['a1', 'a1', 'a2'],
                                      'B': ['b1', 'b2', 'b1'],
                                      'C': ['c1', 'c1', 'c2']})
        >>> estimator = ParameterEstimator(model, data)
        >>> estimator.state_counts('A')
            A
        a1  2
        a2  1
        >>> estimator.state_counts('C')
        A  a1      a2
        B  b1  b2  b1  b2
        C
        c1  1   1   0   0
        c2  0   0   1   0
        """

        parents = sorted(self.model.get_parents(variable))
        return super(ParameterEstimator, self).state_counts(
            variable, parents=parents, weighted=weighted, **kwargs
        )


class StructureEstimator(BaseEstimator):
    def __init__(self, data=None, independencies=None, **kwargs):
        """
        Base class for structure estimators in pgmpy.

        Parameters
        ----------
        data: pandas DataFrame object
            dataframe object where each column represents one variable.
            (If some values in the data are missing the data cells should be set to `numpy.NaN`.
            Note that pandas converts each column containing `numpy.NaN`s to dtype `float`.)

        state_names: dict (optional)
            A dict indicating, for each variable, the discrete set of states (or values)
            that the variable can take. If unspecified, the observed values in the data set
            are taken to be the only possible states.

        complete_samples_only: bool (optional, default `True`)
            Specifies how to deal with missing data, if present. If set to `True` all rows
            that contain `np.Nan` somewhere are ignored. If `False` then, for each variable,
            every row where neither the variable nor its parents are `np.NaN` is used.
            This sets the behavior of the `state_count`-method.
        """

        self.independencies = independencies
        if self.independencies is not None:
            self.variables = self.independencies.get_all_variables()

        super(StructureEstimator, self).__init__(data=data, **kwargs)

    def estimate(self):
        pass
